import java.util.concurrent.Semaphore;
import java.util.*;

/**
 * Benchmark #24 - parallel sorting in main memory.
 *
 * @author  $Author: pepca $
 * @version 1.013 $Rev: 92 $ $Date: 2007-03-14 11:45:45 +0100 (st, 14 III 2007) $
 * @since   1.013
 */
public class BenchmarkParallelSort extends Benchmark
{
  /**
   * Worker task specification.
   */
  private static class WorkerTask
  {
    /** 0 .. initial quick-sort (only {@link #dest} is used), 1 .. merge. */
    public int type;

    /** Source cell #1 (cell-index). */
    public int src1;

    /** Source cell #2 (cell-index). */
    public int src2;

    /** Destination cell index. */
    public int dest;

    /** Destination data size in array items. */
    public int size;

    /** Log(size) = 1 .. finished initial sorting (CELL_SIZE), .. */
    public byte logSize;
  }

  /** Maximal number of worker threads. */
  protected int maxThreads;

  /** Total number of items to be sorted. */
  protected final static int TOTAL_SIZE = 128 << 20;
  //protected final static int TOTAL_SIZE = 32 << 20;

  /** Cell size - the smallest bucket sorted by quick-sort. */
  protected final static int CELL_SIZE = 1 << 15;

  /** Number of cells (equivalent to {@link #TOTAL_SIZE}, the actual array is double-sized). */
  protected final static int CELLS = TOTAL_SIZE / CELL_SIZE;

  /** Array to be sorted (double sized!). */
  protected double[] data;

  /** Which cells are finished? (log(size), double sized!) */
  protected byte[] finish;

  /** Sorting was successful. */
  protected boolean ok;

  /** false .. all worker threads should shut down. */
  protected boolean running;

  /** Finish-signaling object. */
  protected Semaphore fin;

  /** Semaphore to control access to worker tasks (signal {@code =>} a task is ready). */
  protected Semaphore avail;

  /** Task queue (size: {@link #CELLS}). */
  protected WorkerTask[] tasks;

  /** Put index (item index to be putted next time). */
  protected int putPtr;

  /** Get index (item index to be looked at next time). */
  protected int getPtr;

  /** Puts new computing task in the processing queue. */
  protected synchronized void newTask ( WorkerTask task )
  {
      // I'm sure there is free space in the tasks[] array..
    tasks[putPtr++] = task;
    if ( putPtr >= CELLS ) putPtr = 0;

      // release another task..
    avail.release();
  }

  /** Initial-sort task has been finished .. should we start new merge task? */
  protected synchronized void finishSort ( WorkerTask task )
  {
      // check my sibling..
    finish[ task.dest ] = 1;
    int sibling = (task.dest ^ 1);
    if ( finish[ sibling ] != 1 ) return;

      // prepare new merge-task:
    sibling &= -2;
    task.type = 1;
    task.src1 = sibling;
    task.src2 = sibling + 1;
    task.dest = sibling + CELLS;
    task.size = CELL_SIZE + CELL_SIZE;
    task.logSize++;

    newTask( task );
  }

  /** Merge task has been finished .. should we start new (bigger) one? */
  protected synchronized void finishMerge ( WorkerTask task )
  {
      // was I the last merge-task?
    int size = task.size;
    if ( size == TOTAL_SIZE )
    {
      fin.release();
      return;
    }

      // check if my sibling is ready to merge with me..
    finish[ task.dest ] = task.logSize;
    int cells = size / CELL_SIZE;
    int sibling = (task.dest ^ cells);

    if ( finish[ sibling ] != task.logSize )
      return;

      // prepare new merge-task:
    if ( task.dest < sibling ) sibling = task.dest;
    task.src1 = sibling;
    task.src2 = sibling + cells;
    task.dest = (sibling ^ CELLS);
    task.size += task.size;
    task.logSize++;

    newTask( task );
  }

  protected synchronized WorkerTask getTask ()
  {
      // I'm sure there is at least one task available..
    WorkerTask tmp = tasks[getPtr];
    tasks[getPtr++] = null;
    if ( getPtr >= CELLS ) getPtr = 0;

    return tmp;
  }

  /**
   * Worker thread. Able to compute 2 different tasks:
   * 1. initial quick-sort of the data cell,
   * 2. merge two pre-sorted cell segments.
   */
  protected class Worker extends Thread
  {
    public void run ()
    {
      do
      {
          // wait for a task to process..
        try
        {
          avail.acquire();
        }
        catch ( InterruptedException e )
        {}
        if ( !running ) return;

        WorkerTask task = getTask();

        if ( task.type == 0 )               // initial sort
        {
          QuickSortDouble.sort( data, task.dest * CELL_SIZE, CELL_SIZE );
          finishSort( task );
        }

        else                                // merge two already sorted segments
        {
          int in1  = task.src1 * CELL_SIZE;
          int in2  = task.src2 * CELL_SIZE;
          int dst  = task.dest * CELL_SIZE;
          int cnt1 = task.size >> 1;
          int cnt2 = cnt1;

            // merge-loop:
          do
            if ( data[in1] < data[in2] )
            {
              data[dst++] = data[in1++];
              if ( --cnt1 <= 0 ) break;
            }
            else
            {
              data[dst++] = data[in2++];
              if ( --cnt2 <= 0 ) break;
            }
          while ( true );

            // append tail_1?
          while ( cnt1-- > 0 )
            data[dst++] = data[in1++];

            // append tail_2?
          while ( cnt2-- > 0 )
            data[dst++] = data[in2++];

          finishMerge( task );
        }

      } while ( true );
    }
  }

  /**
   * Initializing constructor.
   *
   * @param  thr Max. number of threads to use (should be power of two).
   */
  public BenchmarkParallelSort ( int threads )
  {
    maxThreads = 1;
    while ( maxThreads + maxThreads <= threads )
      maxThreads += maxThreads;

    finish = new byte[ CELLS + CELLS ];
    data  = new double[ TOTAL_SIZE + TOTAL_SIZE ];
  }

  /**
   * Test (benchmark) ordinal number.
   *
   * @return Ordinal number of the test.
   */
  public int testNumber ()
  {
    return 24;
  }

  /**
   * One run (the complete sorting job utilizing the given
   * number of worker threads).
   */
  protected long oneRun ( int threads )
  {
    long startTime = System.currentTimeMillis();

    ok      = false;
    running = true;
    avail   = new Semaphore( 0 );
    fin     = new Semaphore( 0 );
    tasks   = new WorkerTask[ CELLS ];
    putPtr  =
    getPtr  = 0;
    Arrays.fill( finish, (byte)0 );

      // 1. create worker threads:
    Worker[] workers = new Worker[ threads ];
    int i;
    for ( i = 0; i < threads; i++ )
    {
      workers[i] = new Worker();
      workers[i].start();
    }

      // 2. generate primary tasks:
    RandomJames rnd = new RandomJames();
    i = 0;
    int c = 0;
    while ( c < CELLS )
    {
      for ( int j = 0; j++ < CELL_SIZE; )
        data[i++] = rnd.uniformNumber();

      WorkerTask t = new WorkerTask();
      t.type       = 0;
      t.dest       = c;
      t.size       = CELL_SIZE;
      t.logSize    = 1;

      newTask( t );

      c++;
    }

      // 3. wait until everything is finished:
    try
    {
      fin.acquire( 1 );
    }
    catch ( InterruptedException e )
    {}
    running = false;
    avail.release( threads );               // stop all working threads..

    for ( i = 0; i < threads; i++ )
    {
      try
      {
        workers[i].join();
      }
      catch ( InterruptedException e )
      {}
      workers[i] = null;
    }

    fin     = null;
    avail   = null;
    tasks   = null;                         // to be sure

      // 4. check the result:
    for ( i = 1; i < TOTAL_SIZE; i++ )
      if ( data[i-1] > data[i] )
        return -1L;

    ok      = true;
    return( System.currentTimeMillis() - startTime );
  }

  /**
   * The benchmark algorithm itself. Needs not be re-entrant but can be iterated.
   */
  protected void test ()
  {
    boolean totalOk = true;
    int threads = 2;

    do
    {
      long runElapsed = oneRun( threads );
      if ( runElapsed < 0L )
      {
        System.out.println( "Threads " + threads + " failed!" );
        totalOk = false;
      }
      else
      {
        System.out.println( "Threads " + threads + ": " + Bench.formatSec( runElapsed ) + " sec " );
      }
    }
    while ( (threads += threads) <= maxThreads );

    ok = totalOk;
  }

  /**
   * Result string of the test. Should be called after the <code>runTest()</code>
   * was finished.
   *
   * @return Result of the test.
   */
  public String testResult ()
  {
    return( "Parallel sort in virtual memory: " + (ok ? "OK" : "failed") +
            ", double[" + (TOTAL_SIZE >> 20) + "M]" );
  }

}
