import java.util.concurrent.Semaphore;
import java.util.*;

/**
 * Benchmark #22 - parallel SHA-1 computing.
 *
 * @author  $Author: pepca $
 * @version 1.013 $Rev: 93 $ $Date: 2007-03-17 16:25:56 +0100 (so, 17 III 2007) $
 * @since   1.010
 */
public class BenchmarkThread extends Benchmark
{
  /**
   * Worker task specification.
   */
  private static class WorkerTask
  {
    /** 0 .. generate data file, 1 .. SHA-1 digest */
    public int type;

    /** Pseudo-random seed for task #0 */
    public int seed;

    /** Data array for task #1 */
    public byte[] data;
  }

  /** Ordinal number of the test. */
  protected int order;

  /** Actual: data file size. */
  protected final int dataSize;

  /** Actual: number of data files. */
  protected final int dataFiles;

  /** Default (#22): data file size. */
  protected final static int DATA_SIZE = 1 << 20;

  /** Default (#22): number of data files to process. */
  protected final static int DATA_FILES = 4000;

  /** Size of task queue. */
  protected final static int QUEUE_SIZE = 64;

  /** Number of working hreads (#22). */
  protected final static int MAX_THREADS = 16;

  /** Static number of working threads for #23. */
  protected final static int THREADS_23 = 2;

  /** Digest was successful. */
  protected boolean ok;

  /** Computed total digest sum (for checking purpose only). */
  protected final byte[] totalDigest;

  /** Correct total digest sum (#22). */
  protected final static byte[] OK_DIGEST =
  {
    (byte)0xB3, (byte)0x80, (byte)0x98, (byte)0x26,
    (byte)0x3E, (byte)0x74, (byte)0x9B, (byte)0x74,
    (byte)0xAF, (byte)0x6A, (byte)0x36, (byte)0xB6,
    (byte)0xED, (byte)0x1D, (byte)0x5C, (byte)0xDD,
    (byte)0x69, (byte)0xBD, (byte)0x3F, (byte)0x57
  };

  /** Correct total digest sum (#23). */
  protected final static byte[] OK_DIGEST2 =
  {
    (byte)0xC4, (byte)0xF3, (byte)0xF6, (byte)0x51,
    (byte)0x0A, (byte)0x3D, (byte)0x76, (byte)0x5A,
    (byte)0x16, (byte)0xAC, (byte)0x37, (byte)0xB9,
    (byte)0x01, (byte)0xE9, (byte)0x1B, (byte)0x39,
    (byte)0xD8, (byte)0xB4, (byte)0x1E, (byte)0x85
  };

  /** Actual correct digest sum. */
  protected final byte[] okDigest;

  /** Semaphore to control access to worker tasks. */
  protected Semaphore avail;

  /** Flag to stop worker threads. */
  protected boolean running;

  /** Task queue semaphore. */
  protected Semaphore queue;

  /** Task queue. */
  protected WorkerTask[] tasks;

  /** Put index (item index to be putted next time). */
  protected int putPtr = 0;

  /** Get index (item index to be looked at next time). */
  protected int getPtr = 0;

  /**
   * Worker thread. Able to compute 2 different tasks:
   * 1. prepare pseudo-random data file of the given size,
   * 2. compute SHA-1 digest of the given data.
   */
  protected class Worker extends Thread
  {
    /** Improves effectivity.. */
    protected byte[] digest = new byte[ 20 ];

    /** SHA-1 digest algorithm. */
    protected SHA1 sha = new SHA1();

    public void run ()
    {
      do
      {
        try
        {
          avail.acquire();
        }
        catch ( InterruptedException e )
        {}
        if ( !running ) return;

        WorkerTask task = getTask();
        int i;

        if ( task.type == 0 )               // generate data file
        {
          int seed = task.seed;
          task.data = new byte[ dataSize ];
          for ( i = 0; i < dataSize; )
          {
            task.data[i++] = (byte)( (seed >> 19) & 0xff );
            seed = seed * 149 + 10007;
          }
          task.type = 1;
          putTask( task );
          avail.release();                  // process another task
        }

        else                                // compute SHA-1 digest
        {
          sha.init();
          sha.update( task.data );
          setResult( sha.finish() );
          queue.release();                  // making room for another task
        }

      } while ( true );
    }
  }

  /** Worker threads. */
  protected Worker[] workers;

  /**
   * Initializing constructor.
   */
  public BenchmarkThread ( int ord )
  {
    order       = ord;
    totalDigest = new byte[ 20 ];

    if ( order == 22 )
    {
      dataSize  = DATA_SIZE;
      dataFiles = DATA_FILES;
      okDigest  = OK_DIGEST;
    }
    else
    {
      dataSize  = DATA_SIZE >> 13;
      dataFiles = DATA_FILES << 13;
      okDigest  = OK_DIGEST2;
    }
  }

  /**
   * Test (benchmark) ordinal number.
   *
   * @return Ordinal number of the test.
   */
  public int testNumber ()
  {
    return order;
  }

  protected synchronized void putTask ( WorkerTask task )
  {
      // I'm sure there is free space in the tasks[] array..
    tasks[putPtr++] = task;
    if ( putPtr >= QUEUE_SIZE ) putPtr = 0;
  }

  protected synchronized WorkerTask getTask ()
  {
      // I'm sure there is at least one task available..
    WorkerTask tmp = tasks[getPtr];
    tasks[getPtr++] = null;
    if ( getPtr >= QUEUE_SIZE ) getPtr = 0;
    return tmp;
  }

  protected synchronized void setResult ( byte[] digest )
  {
    int i;
    for ( i = 0; i < 20; i++ )
      totalDigest[i] ^= digest[i];
  }

  /**
   * One run (the complete SHA-1 computation utilizing the given
   * number of worker threads).
   */
  protected long oneRun ( int threads )
  {
    long startTime = System.currentTimeMillis();

    running = true;
    Arrays.fill( totalDigest, (byte)0 );
    avail = new Semaphore( 0 );
    tasks = new WorkerTask[ QUEUE_SIZE ];
    queue = new Semaphore( QUEUE_SIZE );

      // 1. create worker threads:
    workers = new Worker[ threads ];
    int i;
    for ( i = 0; i < threads; i++ )
    {
      workers[i] = new Worker();
      workers[i].start();
    }

      // 2. generate primary tasks:
    RandomJames rnd = new RandomJames();
    for ( i = 0; i < dataFiles; i++ )
    {
      try
      {
        queue.acquire();
      }
      catch ( InterruptedException e )
      {}
      WorkerTask t = new WorkerTask();
      t.type = 0;
      t.seed = (int)( rnd.uniformNumber() * 0x7fffffff );
      putTask( t );
      avail.release();
    }

      // 3. wait until everything is finished:
    try
    {
      queue.acquire( QUEUE_SIZE );
    }
    catch ( InterruptedException e )
    {}
    running = false;
    avail.release( threads );               // stop all working threads..

    for ( i = 0; i < threads; i++ )
    {
      try
      {
        workers[i].join();
      }
      catch ( InterruptedException e )
      {}
      workers[i] = null;
    }

      // 4. check the result:
    ok = true;
    for ( i = 0; i < 20; i++ )
      if ( totalDigest[i] != okDigest[i] )
      {
        ok = false;
        break;
      }

    tasks   = null;
    queue   = null;
    workers = null;

    long stopTime = System.currentTimeMillis();
    return( ok ? (stopTime - startTime) : -1L );
  }

  /**
   * The benchmark algorithm itself. Needs not be re-entrant but can be iterated.
   */
  protected void test ()
  {
    if ( order == 22 )
    {
      boolean totalOk = true;
      int threads = 1;

      do
      {
        long runElapsed = oneRun( threads );
        if ( runElapsed < 0L )
        {
          System.out.println( "Threads " + threads + " failed!" );
          totalOk = false;
        }
        else
        {
          System.out.println( "Threads " + threads + ": " + Bench.formatSec( runElapsed ) + " sec " );
        }
      }
      while ( (threads += threads) <= MAX_THREADS );

      ok = totalOk;
    }

    else
     ok = (oneRun( THREADS_23 ) >= 0L);
  }

  /**
   * Result string of the test. Should be called after the <code>runTest()</code>
   * was finished.
   *
   * @return Result of the test.
   */
  public String testResult ()
  {
    StringBuffer str = new StringBuffer( "Parallel SHA-1 digest: files = " );

    if ( order == 22 )
      str.append( dataFiles )
         .append( ", size = " )
         .append( (dataSize >> 10) )
         .append( " KB, threads = " )
         .append( MAX_THREADS );
    else
      str.append( (dataFiles >> 20) )
         .append( " M, size = " )
         .append( dataSize )
         .append( ", threads = " )
         .append( THREADS_23 );

    str.append( ", result = " )
       .append( (ok ? "OK" : "fail") )
       .append( "\n" )
       .append( SHA1.digestToString( totalDigest ) );

    return str.toString();
  }

}
