/**
 *  Benchmark #2 to 3 - Image transposition.
 *   @author  $Author: pepca $
 *   @version 0.009 $Rev: 11 $ $Date: 2004-11-14 22:48:54 +0100 (ne, 14 XI 2004) $
 *   @since   0.007
 */
public class BenchmarkTranspose extends Benchmark {

    /**
     *  Initializing constructor.
     *   @param  ord Ordinal number of the test (valid values: 2 to 3).
     *   @param  _alt Alternative flag.
     */
    public BenchmarkTranspose ( int ord, boolean _correct, boolean _alt ) {
        order = ord;
        correct = _correct;
        arrSize = correct ? order - 17 : order - 2;
        alt = _alt;
    }

    /** Ordinal number of the test. */
    protected final int order;

    /** Array size index */
    protected final int arrSize;

    /** Image sizes (1 byte per pixel). */
    protected final static int size[] = { 4096, 8192 };

    /** Number of iterations. */
    protected final static int iter[] = { 48, 12 };

    /** Test alternative: use byte array access only. */
    protected boolean alt;

    /** Two alternatives as in C++ version */
    protected boolean correct;

    /**
     *  Test (benchmark) ordinal number.
     *   @return Ordinal number of the test.
     */
    public int testNumber () {
        return order;
    }

    /**
     *  Transposes the given image <code>src</code>, result will be written to <code>dest</code>.
     *   @param  dest Result image.
     *   @param  src Source image.
     */
    protected static void transpose ( byte[][] dest, byte[][] src ) {
        int height = dest.length;
        int width = dest[0].length;
        int i, j;
        for ( i = 0; i < height; i++ )
            for ( j = 0; j < width; j++ )
                dest[i][j] = src[j][i];
    }

    /**
     *  Transposes the given image <code>src</code>, result will be written to <code>dest</code>.
     *   @param  dest Result image.
     *   @param  src Source image.
     */
    protected static void transposeCorrect ( int[][] dest, int[][] src ) {
        int height = dest.length;
        int width = dest[0].length;
        int i, j, srcI, srcJ;
        for ( i = srcJ = 0; i < height;
              i += 4, srcJ++ ) {            // one image strip
            for ( j = srcI = 0; j < width;
                  j++, srcI += 4 ) {        // one 4x4 rectangle
                int s0 = src[srcI  ][srcJ];
                int s1 = src[srcI+1][srcJ];
                int s2 = src[srcI+2][srcJ];
                int s3 = src[srcI+3][srcJ];
                dest[i  ][j] =  (s0 & 0xFF)            + ((s1 & 0xFF)<<8) +
                               ((s2 & 0xFF)<<16)       + ((s3 & 0xFF)<<24);
                dest[i+1][j] =  ((s0>>8) & 0xFF)       + (((s1>>8) & 0xFF)<<8) +
                               (((s2>>8) & 0xFF)<<16)  + (((s3>>8) & 0xFF)<<24);
                dest[i+2][j] =  ((s0>>16) & 0xFF)      + (((s1>>16) & 0xFF)<<8) +
                               (((s2>>16) & 0xFF)<<16) + (((s3>>16) & 0xFF)<<24);
                dest[i+3][j] =  ((s0>>24) & 0xFF)      + (((s1>>24) & 0xFF)<<8) +
                               (((s2>>24) & 0xFF)<<16) + (((s3>>24) & 0xFF)<<24);
                }
            }
    }

    /**
     *  Transposes the given image <code>src</code>, result will be written to <code>dest</code>.
     *  Exact copy of the C++-algorithm..
     *   @param  dest Result image.
     *   @param  src Source image.
     */
    protected static void transpose ( int[][] dest, int[][] src ) {
        int height = dest.length;
        int width = dest[0].length;
        int i, j, srcI, srcJ;
        int sact, dact1, dact2, dact3, dact4;
        for ( i = srcJ = 0; i < height;
              i += 4, srcJ++ ) {            // one image strip
            for ( j = srcI = 0; j < width;
                  j++, srcI += 4 ) {        // one 4x4 rectangle
                sact  = src[srcI][srcJ];
                dact1 =     sact & 0xFF;
                dact2 =   ((sact >>  8) & 0xFF);
                dact3 =   ((sact >> 16) & 0xFF);
                dact4 =   ((sact >> 24) & 0xFF);
                dact1 |= (( sact        & 0xFF) <<  8);
                dact2 |= (((sact >>  8) & 0xFF) <<  8);
                dact3 |= (((sact >> 16) & 0xFF) <<  8);
                dact4 |= (((sact >> 24) & 0xFF) <<  8);
                dact1 |= (( sact        & 0xFF) << 16);
                dact2 |= (((sact >>  8) & 0xFF) << 16);
                dact3 |= (((sact >> 16) & 0xFF) << 16);
                dact4 |= (((sact >> 24) & 0xFF) << 16);
                dact1 |= (( sact        & 0xFF) << 24);
                dact2 |= (((sact >>  8) & 0xFF) << 24);
                dact3 |= (((sact >> 16) & 0xFF) << 24);
                dact4 |= (((sact >> 24) & 0xFF) << 24);
                dest[i  ][j] = dact1;
                dest[i+1][j] = dact2;
                dest[i+2][j] = dact3;
                dest[i+3][j] = dact4;
                }
            }
    }

    /**
     *  The benchmark algorithm itself. Needs not be re-entrant but can be iterated.
     */
    protected void test () {
        int i;
        if ( alt ) {                        // byte[][] image representation
            byte[][] im1 = new byte[size[arrSize]][size[arrSize]];
            byte[][] im2 = new byte[size[arrSize]][size[arrSize]];
            for ( i = 0; i++ < iter[arrSize]; ) {
                transpose(im2,im1);
                transpose(im1,im2);
                }
            }
        else {                              // int[][] image representation
            int[][] im1 = new int[size[arrSize]][size[arrSize]>>2];
            int[][] im2 = new int[size[arrSize]][size[arrSize]>>2];
            if ( correct )
              for ( i = 0; i++ < iter[arrSize]; ) {
                  transposeCorrect(im2,im1);
                  transposeCorrect(im1,im2);
                  }
            else
              for ( i = 0; i++ < iter[arrSize]; ) {
                  transpose(im2,im1);
                  transpose(im1,im2);
                  }
            }
    }

    /**
     *  Result string of the test. Should be called after the <code>runTest()</code>
     *  was finished.
     *   @return Result of the test.
     */
    public String testResult () {
        return( "Transpose the [" + size[arrSize] + "," + size[arrSize] + "] image" );
    }

}
