/*
    bitmask.cpp

    variable-size bit masks

    $Rev: 8 $
    $Date: 2004-11-07 05:01:56 +0100 (ne, 07 XI 2004) $

    Copyright (C) 2000-2004 by Josef Pelikan, MFF UK Prague
        http://cgg.ms.mff.cuni.cz/~pepca/

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pch.h"

//-----------------------------------------------------------
//  CBitMask object:

CBitMask::CBitMask ()
{
    mask = NULL;
    allocMask = min = max = 0;
}

CBitMask::CBitMask ( const CBitMask *b )
{
    mask = NULL;
    assign(b);
}

CBitMask *CBitMask::clone () const
    // clones the actual object instance
{
    return new CBitMask(this);
}

void CBitMask::assign ( const CBitMask *b )
{
    empty();
    if ( b ) {
        min = b->min;
        max = b->max;
        if ( (allocMask = b->allocMask) ) {
#ifdef USE_64_BITS
            mask = (unsigned64*)VIRTUAL_ALLOC(allocMask);
#else
            mask = (byte*)VIRTUAL_ALLOC(allocMask);
#endif
            memcpy(mask,b->mask,allocMask);
            }
        }
}

void CBitMask::empty ()
    // (re)initializes the mask
{
    if ( mask ) {
        VIRTUAL_FREE(mask,allocMask);
        mask = NULL;
        }
    allocMask = min = max = 0;
}

CBitMask::~CBitMask ()
{
    empty();
}

//-----------------------------------------------------------
//  access:

void CBitMask::assert ( int value )
    // prepares the mask array to hold the given value
{
#ifdef USE_64_BITS
    int add;
    unsigned64 *newMask;
    if ( value < min ) {
        add = (min - value + 63) >> 3;
        newMask = (unsigned64*)VIRTUAL_ALLOC(allocMask+add);
        memset(newMask,0,add);
        memcpy(newMask+(add>>3),mask,allocMask);
        VIRTUAL_FREE(mask,allocMask);
        mask = newMask;
        allocMask += add;
        min -= add << 3;
        return;
        }
    if ( value >= max ) {
        add = (value - min + 64) >> 3;
        newMask = (unsigned64*)VIRTUAL_ALLOC(add);
        memcpy(newMask,mask,allocMask);
        memset(newMask+(allocMask>>3),0,add-allocMask);
        VIRTUAL_FREE(mask,allocMask);
        mask = newMask;
        allocMask = add;
        max = min + (add << 3);
        }
#else
    int add;
    byte *newMask;
    if ( value < min ) {
        add = (min - value + 7) >> 3;
        newMask = (byte*)VIRTUAL_ALLOC(allocMask+add);
        memset(newMask,0,add);
        memcpy(newMask+add,mask,allocMask);
        VIRTUAL_FREE(mask,allocMask);
        mask = newMask;
        allocMask += add;
        min -= add << 3;
        return;
        }
    if ( value >= max ) {
        add = (value - min + 8) >> 3;
        newMask = (byte*)VIRTUAL_ALLOC(add);
        memcpy(newMask,mask,allocMask);
        memset(newMask+allocMask,0,add-allocMask);
        VIRTUAL_FREE(mask,allocMask);
        mask = newMask;
        allocMask = add;
        max = min + (add << 3);
        }
#endif
}

void CBitMask::set ( int value, bool flag )
    // sets one flag value
{
    if ( flag ) {
        assert(value);
        value -= min;
#ifdef USE_64_BITS
        mask[value>>6] |= (unsigned64)1 << (value & 63);
#else
        mask[value>>3] |= 1 << (value & 7);
#endif
        return;
        }
    if ( value < min || value >= max ) return;
    value -= min;
#ifdef USE_64_BITS
    mask[value>>6] &= ~((unsigned64)1 << (value & 63));
#else
    mask[value>>3] &= ~(1 << (value & 7));
#endif
}

void CBitMask::on ( int value )
{
    assert(value);
    value -= min;
#ifdef USE_64_BITS
    mask[value>>6] |= (unsigned64)1 << (value & 63);
#else
    mask[value>>3] |= 1 << (value & 7);
#endif
}

void CBitMask::off ( int value )
{
    if ( value < min || value >= max ) return;
    value -= min;
#ifdef USE_64_BITS
    mask[value>>6] &= ~((unsigned64)1 << (value & 63));
#else
    mask[value>>3] &= ~(1 << (value & 7));
#endif
}

void CBitMask::range ( int from, int len, bool flag )
    // sets the given range
{
    if ( flag ) {
        assert(from);
        assert(from+len-1);
        }
    else {
        if ( from < min ) {
            len -= min - from;
            from = min;
            }
        if ( from + len > max )
            len = max - from;
        }
    while ( len-- > 0 ) setUnsafe(from++,flag);
}

int CBitMask::card () const
    // bit-mask cardinality
{
    if ( !allocMask ) return 0;
    int count = 0;
#ifdef USE_64_BITS
    unsigned64 *pm = mask;
    for ( int i = -8; i += 8 < allocMask; ) {
        unsigned64 act = *pm++;
        while ( act ) {
            count += (int)(act & 1);
            act >>= 1;
            }
        }
#else
    byte *pm = mask;
    for ( int i = 0; i++ < allocMask; ) {
        unsigned act = *pm++;
        while ( act ) {
            count += (act & 1);
            act >>= 1;
            }
        }
#endif
    return count;
}

int CBitMask::getFirst () const
    // returns the first bit or MaskEnd
{
    if ( !mask ) return MaskEnd;
    return getNext(min-1);
}

int CBitMask::getNext ( int i ) const
    // returns the next bit or MaskEnd
{
    int ii;
#ifdef USE_64_BITS
    unsigned64 b;
    while ( ++i < max )
        if ( (b = mask[(ii = i - min) >> 6]) ) {
            if ( b & ((unsigned64)1 << (ii & 63)) ) return i;
            }
        else {
            i += 64 - (ii & 63);
            ii >>= 6;
            while ( i < max && !mask[++ii] ) i += 64;
            i--;
            }
#else
    byte b;
    while ( ++i < max )
        if ( (b = mask[(ii = i - min) >> 3]) ) {
            if ( b & (1<<(ii & 7)) ) return i;
            }
        else {
            i += 8 - (ii & 7);
            ii >>= 3;
            while ( i < max && !mask[++ii] ) i += 8;
            i--;
            }
#endif
    return MaskEnd;
}

int CBitMask::getLast () const
{
    if ( !mask ) return MaskEnd;
#ifdef USE_64_BITS
    int ii = ((max - min) >> 6) - 1;
    while ( ii >= 0 && !mask[ii] ) ii--;
    if ( ii < 0 ) return MaskEnd;
    unsigned64 b = mask[ii];
    ii = (ii << 6) + min + 63;
    while ( !(b & 0x8000000000000000) ) {
        b += b;
        ii--;
        }
#else
    int ii = ((max - min) >> 3) - 1;
    while ( ii >= 0 && !mask[ii] ) ii--;
    if ( ii < 0 ) return MaskEnd;
    byte b = mask[ii];
    ii = (ii << 3) + min + 7;
    while ( !(b & 0x80) ) {
        b += b;
        ii--;
        }
#endif
    return ii;
}

//-----------------------------------------------------------
//  set operations were removed in 1.008
