/*
    main.cc: VUCAKO Bench (CPU/memory/cache/disk benchmark).

    $Rev: 93 $
    $Date: 2007-03-17 16:25:56 +0100 (so, 17 III 2007) $

    Copyright (C) 2001-2007 by Josef Pelikan, MFF UK Prague
        http://cgg.ms.mff.cuni.cz/~pepca/

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pch.h"

//-----------------------------------------------------------
//  state variables:

bool verbose     = true;                    // verbose listing mode
bool superBrief  = false;                   // "super-brief" listing mode
bool changed     = false;                   // results were changed
bool sort        = false;                   // sort the result table
bool ascending   = true;                    // ascending order
bool listing     = false;                   // listing mode
bool spreadSheet = false;                   // spread-sheet readable mode
bool logAll      = false;                   // log all results (for statistics purposes)
int  completion;                            // return code
int  width       = 80;                      // width of the listing

#define SortTime        0                   // time in seconds
#define SortRef         1                   // ref number (time/clocks)
#define SortClock       2                   // CPU clock speed in MHz

int sortType = SortTime;

unsigned restrictMemory  = 0;
unsigned restrictVirtual = 0;
unsigned restrictDisk    = 0;

VRegularExpression *restrictExpr = NULL;    // restriction of result table

int enhancedCount = 0;                      // number of enhanced result items (merge)

//-----------------------------------------------------------
//  data persistence:

#define DataFile        "results.txt"
#define BackupFile      "results.bak"
#define EnhancedFile    "enhanced.txt"
#define InitData        128
#define MaxProcName     128
#define CPUFile         "cpu.txt"
#define UndefinedCPU    "undefined"
#define UndefinedClock  1
#define SleepFlush      20000

struct DataItem
{
  char *procName;                           // these strings are shared (on the heap)
  int clock;                                // CPU clock speed in MHz
  int stat;                                 // test ordinal number
  int time;                                 // time in milliseconds
  bool enhanced;                            // was this result enhanced?
};

DataItem *di = NULL;
int diLen    = 0;
int diAlloc  = 0;

char *actCpu = UndefinedCPU;
char cpuBuf[ MaxProcName ];
int actClock = 0;

void assertData ( int i )
{
  if ( i <= diAlloc ) return;

  i = (i + InitData-1) & -InitData;
  DataItem *ndi = new DataItem[ i ];
  if ( diAlloc )
  {
    memcpy( ndi, di, diAlloc * sizeof(DataItem) );
    delete []di;
  }
  di = ndi;
  diAlloc = i;
}

char *assignProcName ( const char *name )
{
  int i;
  for ( i = 0; i < diLen; i++ )
    if ( !strcmp( name, di[i].procName ) ) return di[i].procName;

  char *nn = new char[ strlen(name)+1 ];
  strcpy( nn, name );
  return nn;
}

bool readData ( const char *fileName =NULL )
    // <test#> <time-in-millisec> <clock-MHz> <CPU-name>
{
  if ( !fileName ) fileName = DataFile;
  char buf[ MaxProcName<<1 ];
  FILE *f = fopen( fileName, "rt" );
  if ( !f ) return false;

  do
  {
    if ( !fgets( buf, MaxProcName<<1, f ) ) break;
    char *ptr = buf;
    int  stat = 0;
    int read;
    if ( sscanf( ptr, "%d%n", &stat, &read ) < 1 || stat < 1 ) break;
    ptr += read;
    while ( buf[strlen( buf )-1] == '\n' || isspace( buf[strlen( buf )-1] ) )
      buf[strlen( buf )-1] = (char)0;
    int ti, cl;
    if ( sscanf( ptr, "%d%d%n", &ti, &cl, &read ) < 2 ) break;
    ptr += read;
    assertData( diLen+1 );
    di[diLen].stat = stat;
    di[diLen].time = ti;
    di[diLen].clock = cl;
    while ( isspace( *ptr ) ) ptr++;
    di[diLen].procName = assignProcName( ptr );
    di[diLen].enhanced = false;
    diLen++;
  } while ( true );

  fclose( f );
  return true;
}

void unifyData ()
{
    int i, j;
    changed = true;
    for ( i = 0; i < diLen; i++ )
        if ( !di[i].enhanced ) {
            int clock = di[i].clock;
            const char *procName = di[i].procName;
            int stat = di[i].stat;
            for ( j = i + 1; j < diLen; j++ )
                if ( !di[j].enhanced && clock == di[j].clock &&
                     procName == di[j].procName && stat == di[j].stat ) {
                    di[j].enhanced = true;
                    if ( di[j].time < di[i].time ) {
                        enhancedCount++;
                        di[i].time = di[j].time;
                        }
                    }
            }
        // compress the array:
    i = 0;
    for ( j = 0; j < diLen; j++ )
        if ( !di[j].enhanced ) {
            if ( i < j ) memcpy(di+i,di+j,sizeof(DataItem));
            i++;
            }
    diLen = i;
}

bool writeData ()
{
    if ( !changed ) return false;
    FILE *f = fopen(DataFile,"wt");
    FILE *fe = fopen(EnhancedFile,"a+t");
    if ( !f || !fe ) return false;
    int i;
    for ( i = 0; i < diLen; i++ ) {
        fprintf(f,"%2d %8d %4d %s\n",di[i].stat,di[i].time,di[i].clock,di[i].procName);
        if ( di[i].enhanced ) {
            fprintf(fe,"%2d %8d %4d %s\n",di[i].stat,di[i].time,di[i].clock,di[i].procName);
            di[i].enhanced = false;
            }
        }
    fclose(f);
    fclose(fe);
    changed = false;
    return true;
}

void newResult ( int stat, int elapsed )
{
    char buf[64];
    int i;
    bool insert = true;                     // this test was run for the first time
    int old;                                // old result

        // print enhancement first:
    for ( i = 0; i < diLen; i++ )
        if ( di[i].stat == stat &&
            !strcmp(di[i].procName,actCpu) &&
            di[i].clock == actClock ) {     // the same hardware
                insert = false;
                old = di[i].time;
                if ( elapsed < old ) {
                    di[i].time = elapsed;
                    di[i].enhanced = changed = true;
                    }
                }
    strncpy(buf,actCpu,60);

        // verbose output:
    if ( verbose ) {
        buf[52] = (char)0;
        printf( "%-52s%7.2fs",buf,elapsed*0.001);
        if ( insert )
            printf( "   NEW");
        else
            printf( "%+6.2f",(elapsed-old)*0.001);
        printf( " ref:%8.2f\n",elapsed*1.e-6*actClock);
            // print comparisons:
        for ( i = 0; i < diLen; i++ )
            if ( di[i].stat == stat &&
                 (strcmp(di[i].procName,actCpu) || di[i].clock != actClock) &&
                 (!restrictExpr || restrictExpr->match(di[i].procName)) ) {
                    // compare it
                strncpy(buf,di[i].procName,52);
                buf[52] = (char)0;
                printf( "%-52s%7.2fs%5.0f%% ref:%8.2f\n",
                       buf,di[i].time*0.001,(elapsed*100.0)/di[i].time,
                       di[i].time*1.e-6*di[i].clock);
                }
        }

        // brief output:
    else {
        if ( superBrief ) {                 // print test ordinal # in super-brief mode
            buf[58] = (char)0;
            printf( "%2d %-58s",stat,buf);
            }
        else {
            buf[61] = (char)0;
            printf( "%-61s",buf);
            }
        printf( "%9.3fs",elapsed*0.001);
        if ( insert )
            printf( "     NEW\n");
        else
            printf( "%7.2f%%\n",(elapsed*100.0)/old);
        }

        // insert the new result:
    if ( insert ) {
        changed = true;
        assertData(diLen+1);
        di[diLen].stat = stat;
        di[diLen].time = elapsed;
        di[diLen].clock = actClock;
        di[diLen].procName = assignProcName(actCpu);
        di[diLen].enhanced = true;
        diLen++;
        }

        // log all results?
    if ( logAll ) {
        sprintf(buf,"%d.txt",stat);
        FILE *fa = fopen(buf,"a+t");
        if ( fa )
            fprintf(fa,"%8d %4d %s\n",elapsed,actClock,actCpu);
        fclose(fa);
        }

        // sync the data:
    writeData();
    SLEEP(SleepFlush);
}

int __CDECL cmpTime ( const void *e1, const void *e2 )
{
    DataItem *di1 = di + *((int*)e1);
    DataItem *di2 = di + *((int*)e2);
    int t1 = di1->time;
    int t2 = di2->time;

    if ( t1 == t2 )
    {
      double ref1 = di1->time * 1e-6 * di1->clock;
      double ref2 = di2->time * 1e-6 * di2->clock;

      if ( ref1 == ref2 )
      {
        int names = strcmp( di1->procName, di2->procName );
        return( ascending ? names : -names );
      }
      return( (ref1 < ref2) == ascending ? -1 : 1 );
    }

    return( (t1 < t2) == ascending ? -1 : 1 );
}

int __CDECL cmpRef ( const void *e1, const void *e2 )
{
    DataItem *di1 = di + *((int*)e1);
    DataItem *di2 = di + *((int*)e2);
    double ref1 = di1->time * 1e-6 * di1->clock;
    double ref2 = di2->time * 1e-6 * di2->clock;

    if ( ref1 == ref2 )
    {
      int t1 = di1->time;
      int t2 = di2->time;

      if ( t1 == t2 )
      {
        int names = strcmp( di1->procName, di2->procName );
        return( ascending ? names : -names );
      }
      return( (t1 < t2) == ascending ? -1 : 1 );
    }

    return( (ref1 < ref2) == ascending ? -1 : 1 );
}

int __CDECL cmpClock ( const void *e1, const void *e2 )
{
    DataItem *di1 = di + *((int*)e1);
    DataItem *di2 = di + *((int*)e2);
    int t1 = di1->clock;
    int t2 = di2->clock;

    if ( t1 == t2 )
    {
      int names = strcmp( di1->procName, di2->procName );
      return( ascending ? names : -names );
    }

    return( (t1 < t2) == ascending ? -1 : 1 );
}

void listResult ( int stat )
{
  printf( spreadSheet ? "\n\"  " : "\n  %d: ", stat );

  switch ( stat )
  {
    case  1:
      printf( "Eratosthen (%d-th prime number):",
              PrimeOrder );
      break;
    case  2:
      printf( "Flawed transposition of %dMB image:",
              (ImageWidth1*ImageHeight1)>>20 );
      break;
    case  3:
      printf( "Flawed transposition of %dMB image:",
              (ImageWidth2*ImageHeight2)>>20 );
      break;
    case  4:
      printf( "Arbitrage sequence lookup (double[%d][%d] matrix):",
              ArbitrageN, ArbitrageN );
      break;
    case  5:
      printf( "Needle-throwing simulation (Monte-Carlo, %dM iterations):",
              NeedleMax/1000000 );
      break;
    case  6:
      printf( "Fast memory-copy (two %d MB arrays, %d iterations):",
              MemCopyMax>>20, MemCopyIter );
      break;
    case  7:
      printf( "Monte-Carlo form-factor (%dM iterations):",
              FFIter/1000000 );
      break;
    case  8:
      printf( "Merge-sort on disk (double[%dM] array):",
              SortSize>>20 );
      break;
    case  9:
      printf( "Wavelet (T-S) transform in memory (int32[%dM-%dK] arrays):",
              WaveMaxSize>>20, WaveMinSize>>10 );
      break;
    case 10:
      printf( "Wavelet (T-S) transform on disk (double[%dM-%dK] arrays):",
              WaveDMaxSize>>20, WaveDMinSize>>10 );
      break;
    case 11:
      printf( "Adaptive K-D tree on disk (%dK operations):",
              TreeOperations/1000 );
      break;
    case 12:
      printf( "Quick-sort in virtual memory (double[%dM] array):",
              QuickSortLen1>>20 );
      break;
    case 13:
      printf( "Quick-sort in virtual memory (double[%dM] array):",
              QuickSortLen2>>20 );
      break;
    case 14:
      printf( "Quick-sort in virtual memory (double[%dM] array):",
              QuickSortLen3>>20 );
      break;
    case 15:
      printf( "Garbage collection test in 16MB of VM:" );
      break;
    case 16:
      printf( "Regular expression test (4 x %dk matches on %d-byte strings):",
              RegIter/1000, RegLen );
      break;
    case 17:
      printf( "Transposition of %dMB image:",
              (ImageWidth1*ImageHeight1)>>20 );
      break;
    case 18:
      printf( "Transposition of %dMB image:",
              (ImageWidth2*ImageHeight2)>>20 );
      break;
    case 19:
      printf( "Merge-sort on disk (double[%dM] array):",
              SortSizeBig>>20 );
      break;
    case 20:
      printf( "Wavelet (T-S) transform on disk (double[%dM-%dK] arrays):",
              WaveDMaxSizeBig>>20, WaveDMinSizeBig>>10 );
      break;
    case 21:
      printf( "Adaptive K-D tree on disk (%dK operations):",
              TreeOperationsBig/1000 );
      break;
    case 22:
      printf( "Parallel SHA-1 digest (1 to %d threads, %dx%dMB total):",
              ShaMaxThreads, ShaDataFiles, ShaDataSize>>20 );
      break;
    case 23:
      printf( "Ctx-switch efficiency - SHA-1 digest (%d threads, %dMx%d total):",
              ShaThreads23, (ShaDataFiles<<ShaShift23)>>20, ShaDataSize>>ShaShift23 );
      break;
    case 24:
      printf( "Parallel sort in virtual memory (2 to %d threads, 2x double[%dM] array):",
              PsMaxThreads, ParallelSortLen>>20 );
      break;
    }
    printf( spreadSheet ? "\",Clock[MHz],Time[s],Ref\n" : "\n" );

    // print comparisons:
  int i;
  int *outTable = new int[ diLen ];         // output table to be sorted
  int outLen = 0;                           // length of output table
  for ( i = 0; i < diLen; i++ )
    if ( di[i].stat == stat &&
         (!restrictExpr || restrictExpr->match( di[i].procName )) )
      outTable[outLen++] = i;

    // sort the table:
  if ( sort )
    switch ( sortType )
    {
      case SortTime:
        qsort( outTable, outLen, sizeof(int*), &cmpTime );
        break;
      case SortRef:
        qsort( outTable, outLen, sizeof(int*), &cmpRef );
        break;
      case SortClock:
        qsort( outTable, outLen, sizeof(int*), &cmpClock );
        break;
    }

    // print the table:
  if ( spreadSheet )                        // spread-sheet readable format (CSV)
  {
    for ( i = 0; i < outLen; i++ )
      printf( "\"%s\",%d,%.2f,%.2f\n",
              di[outTable[i]].procName,
              di[outTable[i]].clock,
              di[outTable[i]].time*0.001,
              di[outTable[i]].time*1.e-6*di[outTable[i]].clock );
  }
  else                                      // plain-text table
  {
    char buf[ 144 ];
    for ( i = 0; i < outLen; i++ )
    {
      strncpy( buf, di[outTable[i]].procName, width-23 );
      buf[width-23] = (char)0;
      printf( "%-*s%8.2fs, ref:%7.2f\n",
              width-23,buf,
              di[outTable[i]].time*0.001,
              di[outTable[i]].time*1.e-6*di[outTable[i]].clock );
    }
  }

  delete []outTable;
}

//-----------------------------------------------------------
//  test suite globals:

#define MaxTest         24
#define MaxTestMask     ((1<<MaxTest)-1)
#define DefaultTestMask MaxTestMask

#define Mega(x)         ((x)<<20)
#define Giga(x)         ((x)<<30)
#define ReserveMemory   Mega(3)
#define ReserveVirtual  Mega(12)
#define ReserveDisk     Mega(1)

unsigned memoryLimits[] =
{
  Mega(13)  + ReserveMemory,                //  1: Eratosthen
  Mega(32)  + ReserveMemory,                //  2: Flawed transposition of 16MB image
  Mega(128) + ReserveMemory,                //  3: Flawed transposition of 64MB image
  Mega(0)   + ReserveMemory,                //  4: Arbitrage
  Mega(0)   + ReserveMemory,                //  5: Needle
  Mega(32)  + ReserveMemory,                //  6: Memory copy
  Mega(0)   + ReserveMemory,                //  7: Monte-Carlo form-factor
  Mega(8)   + ReserveMemory,                //  8: Merge-sort on disk
  Mega(16)  + ReserveMemory,                //  9: Wavelets in memory
  Mega(0)   + ReserveMemory,                // 10: Wavelets on disk
  Mega(8)   + ReserveMemory,                // 11: A-K-D tree on disk
  Mega(0)   + ReserveMemory,                // 12: Quick-sort in VM (64M)
  Mega(0)   + ReserveMemory,                // 13: Quick-sort in VM (256M)
  Mega(0)   + ReserveMemory,                // 14: Quick-sort in VM (1G)
  Mega(16)  + ReserveMemory,                // 15: Garbage collection test
  Mega(1)   + ReserveMemory,                // 16: Regular expression test
  Mega(32)  + ReserveMemory,                // 17: Transposition of 16MB image
  Mega(128) + ReserveMemory,                // 18: Transposition of 64MB image
  Mega(8)   + ReserveMemory,                // 19: Merge-sort on disk (big)
  Mega(0)   + ReserveMemory,                // 20: Wavelets on disk (big)
  Mega(8)   + ReserveMemory,                // 21: A-K-D tree on disk (big)
  Mega(64)  + ReserveMemory,                // 22: Parallel SHA-1 (64M)
  Mega(1)   + ReserveMemory,                // 23: Ctx-switch efficiency (1M)
  Mega(0)   + ReserveMemory,                // 24: Parallel sort in VM (2G)
};

unsigned virtualLimits[] =
{
  Mega(13)  + ReserveVirtual,               //  1: Eratosthen
  Mega(32)  + ReserveVirtual,               //  2: Flawed transposition of 16MB image
  Mega(128) + ReserveVirtual,               //  3: Flawed transposition of 64MB image
  Mega(0)   + ReserveVirtual,               //  4: Arbitrage
  Mega(0)   + ReserveVirtual,               //  5: Needle
  Mega(32)  + ReserveVirtual,               //  6: Memory copy
  Mega(0)   + ReserveVirtual,               //  7: Monte-Carlo form-factor
  Mega(8)   + ReserveVirtual,               //  8: Merge-sort on disk
  Mega(16)  + ReserveVirtual,               //  9: Wavelets in memory
  Mega(0)   + ReserveVirtual,               // 10: Wavelets on disk
  Mega(8)   + ReserveVirtual,               // 11: A-K-D tree on disk
  Mega(64)  + ReserveVirtual,               // 12: Quick-sort in VM (64M)
  Mega(256) + ReserveVirtual,               // 13: Quick-sort in VM (256M)
  Giga(1)   + ReserveVirtual,               // 14: Quick-sort in VM (1G)
  Mega(16)  + ReserveVirtual,               // 15: Garbage collection test
  Mega(1)   + ReserveVirtual,               // 16: Regular expression test
  Mega(32)  + ReserveVirtual,               // 17: Transposition of 16MB image
  Mega(128) + ReserveVirtual,               // 18: Transposition of 64MB image
  Mega(8)   + ReserveVirtual,               // 19: Merge-sort on disk (big)
  Mega(0)   + ReserveVirtual,               // 20: Wavelets on disk (big)
  Mega(8)   + ReserveVirtual,               // 21: A-K-D tree on disk (big)
  Mega(64)  + ReserveVirtual,               // 22: Parallel SHA-1 (64M)
  Mega(1)   + ReserveVirtual,               // 23: Ctx-switch efficiency (1M)
  Giga(2)   + ReserveVirtual,               // 24: Parallel sort in VM (2G)
};

unsigned diskLimits[] =
{
  Mega(0)   + ReserveDisk,                  //  1: Eratosthen
  Mega(0)   + ReserveDisk,                  //  2: Flawed transposition of 16MB image
  Mega(0)   + ReserveDisk,                  //  3: Flawed transposition of 64MB image
  Mega(0)   + ReserveDisk,                  //  4: Arbitrage
  Mega(0)   + ReserveDisk,                  //  5: Needle
  Mega(0)   + ReserveDisk,                  //  6: Memory copy
  Mega(0)   + ReserveDisk,                  //  7: Monte-Carlo form-factor
  Mega(128) + ReserveDisk,                  //  8: Merge-sort on disk
  Mega(0)   + ReserveDisk,                  //  9: Wavelets in memory
  Mega(128) + ReserveDisk,                  // 10: Wavelets on disk
  Mega(58)  + ReserveDisk,                  // 11: A-K-D tree on disk
  Mega(0)   + ReserveDisk,                  // 12: Quick-sort in VM (64M)
  Mega(0)   + ReserveDisk,                  // 13: Quick-sort in VM (256M)
  Mega(0)   + ReserveDisk,                  // 14: Quick-sort in VM (1G)
  Mega(0)   + ReserveDisk,                  // 15: Garbage collection test
  Mega(0)   + ReserveDisk,                  // 16: Regular expression test
  Mega(0)   + ReserveDisk,                  // 17: Transposition of 16MB image
  Mega(0)   + ReserveDisk,                  // 18: Transposition of 64MB image
  Giga(1)   + ReserveDisk,                  // 19: Merge-sort on disk (big)
  Giga(1)   + ReserveDisk,                  // 20: Wavelets on disk (big)
  Mega(567) + ReserveDisk,                  // 21: A-K-D tree on disk (big)
  Mega(0)   + ReserveDisk,                  // 22: Parallel SHA-1 (64M)
  Mega(0)   + ReserveDisk,                  // 23: Ctx-switch efficiency (1M)
  Mega(0)   + ReserveDisk,                  // 24: Parallel sort in VM (2G)
};

unsigned readBytes ( const char *len )
{
  char *ePtr = NULL;
  unsigned long result = strtoul( len, &ePtr, 10 );
  if ( ePtr )
    switch ( ePtr[0] )
    {
      case 'g':
      case 'G':
        result <<= 10;
      case 'm':
      case 'M':
        result <<= 10;
      case 'k':
      case 'K':
        result <<= 10;
    }

  return (unsigned)result;
}

void notEnoughMemory ( int mb )
{
  fprintf( stderr, "%d MB of main memory cannot be allocated!\n", mb );
  completion = 11;
}

//-----------------------------------------------------------
//  main routine:

void printHeader ()
{
  printf( "VUCAKO Bench %s-rev%d (c) 2001-07 Josef.Pelikan@mff.cuni.cz (Open Src)\n"
          "  Please send result file (%s) back to author!\n", VERSION, REVISION, DataFile );
}

void printUsage ()
{
  printHeader();
  printf( "Usage: bench [-h] [-l[s]] [-m <result-file>] [-b] [-s{a|d}{t|r|c}] [-r{m|v|d}]\n"
          "             [-i <test>] [-t] [-r <reg-exp>] [-c <clock> [\"CPU-description\"]]\n"
          "             [<mask>]\n"
          "    -h      print help message (this text)\n"
          "    -l[s|<width>]\n"
          "            don't run any tests, print result table(s) only\n"
          "            \"s\" generates \"spread-sheet readable\" output\n"
          "    -m <result-file> merge the given result file\n"
          "    -b[b]   [super-]brief mode (don't print header and old results)\n"
          "    -s{a|d}{t|r|c}\n"
          "            sort results (ascending/descending, time/ref/clock)\n"
          "    -r{m|v|d} <bytes>\n"
          "            restrict tests according to physical RAM (m),\n"
          "            virtual memory (v) or free disk space (d).\n"
          "            <bytes> is number of bytes, examples: 256k, 64m, 2g\n"
          "    -i <test> run only the given test (<test> - see below)\n"
          "    -t      self-test (pseudo-random generator, SHA-1)\n"
          "    -r <reg-exp> restrict result table to systems matching <reg-exp>\n"
          "    -c      system-description:\n"
          "      <clock> is CPU clock in MHz (see the 2nd column of \"%s\")\n"
          "      \"CPU-description\" is arbitrary string (abbreviations are preferred,\n"
          "                        see \"%s\" ignoring first two columns)\n"
          "    if -c is not specified, the first line of \"%s\" will be read\n"
          "    <mask>  is bit-mask (hexadecimal number 0xNNNN can be used):\n"
          "      1. 0x000001 .. Eratosthen (%d-th prime number)\n"
          "      2. 0x000002 .. Flawed transposition of %dMB image (uses two buffers!)\n"
          "      3. 0x000004 .. Flawed transposition of %dMB image (uses two buffers!)\n"
          "      4. 0x000008 .. Arbitrage sequence lookup (double[%d][%d] matrix)\n"
          "      5. 0x000010 .. Needle-throwing simulation (Monte-Carlo, %dM iterations)\n"
          "      6. 0x000020 .. Fast memory-copy (two %d MB arrays, %d iterations)\n"
          "      7. 0x000040 .. Monte-Carlo form-factor (%dM iterations)\n"
          "      8. 0x000080 .. Merge-sort on disk (double[%dM] array)\n"
          "      9. 0x000100 .. Wavelet (T-S) transform in memory (int32[%dM-%dK] arrays)\n"
          "     10. 0x000200 .. Wavelet (T-S) transform on disk (double[%dM-%dK] arrays)\n"
          "     11. 0x000400 .. Adaptive K-D tree on disk (%dK operations)\n"
          "     12. 0x000800 .. Quick-sort in virtual memory (double[%dM] array)\n"
          "     13. 0x001000 .. Quick-sort in virtual memory (double[%dM] array)\n"
          "     14. 0x002000 .. Quick-sort in virtual memory (double[%dM] array)\n"
          "     15. 0x004000 .. Garbage collection test in 16MB of VM\n"
          "     16. 0x008000 .. Regular expression test (%dk matches on %d-bytes strings)\n"
          "     17. 0x010000 .. Transposition of %dMB image (uses two buffers!)\n"
          "     18. 0x020000 .. Transposition of %dMB image (uses two buffers!)\n"
          "     19. 0x040000 .. Merge-sort on disk (double[%dM] array)\n"
          "     20. 0x080000 .. Wavelet (T-S) transform on disk (double[%dM-%dM] arr)\n"
          "     21. 0x100000 .. Adaptive K-D tree on disk (%dK operations)\n"
          "     22. 0x200000 .. Parallel SHA-1 digest (1 to %d threads, %dx%dMB total)\n"
          "     23. 0x400000 .. Ctx-switch test - SHA-1 (%d threads, %dMx%d total)\n"
          "     24. 0x800000 .. Parallel sort in VM (2 to %d threads, 2x double[%dM] arr)\n"
          , CPUFile, CPUFile, CPUFile,
          PrimeOrder,                                      // test  1
          (ImageWidth1*ImageHeight1)>>20,                  // test  2
          (ImageWidth2*ImageHeight2)>>20,                  // test  3
          ArbitrageN, ArbitrageN,                          // test  4
          NeedleMax/1000000,                               // test  5
          MemCopyMax>>20, MemCopyIter,                     // test  6
          FFIter/1000000,                                  // test  7
          SortSize>>20,                                    // test  8
          WaveMaxSize>>20, WaveMinSize>>10,                // test  9
          WaveDMaxSize>>20, WaveDMinSize>>10,              // test 10
          TreeOperations/1000,                             // test 11
          QuickSortLen1>>20,                               // test 12
          QuickSortLen2>>20,                               // test 13
          QuickSortLen3>>20,                               // test 14
          (RegIter<<2)/1000, RegLen,                       // test 16
          (ImageWidth1*ImageHeight1)>>20,                  // test 17
          (ImageWidth2*ImageHeight2)>>20,                  // test 18
          SortSizeBig>>20,                                 // test 19
          WaveDMaxSizeBig>>20, WaveDMinSizeBig>>20,        // test 20
          TreeOperationsBig/1000,                          // test 21
          ShaMaxThreads, ShaDataFiles, ShaDataSize>>20,    // test 22
          ShaThreads23, (ShaDataFiles<<ShaShift23)>>20,
          ShaDataSize>>ShaShift23,                         // test 23
          PsMaxThreads, ParallelSortLen>>20                // test 24
        );
}

int __CDECL main ( int argc, char **argv )
{
      // compile-time asserts:
    COMPILE_ASSERT( sizeof(byte) == 1 );
    COMPILE_ASSERT( sizeof(int32) == 4 );
    COMPILE_ASSERT( sizeof(unsigned32) == 4 );
#ifdef USE_64_BITS
    COMPILE_ASSERT( sizeof(unsigned64) == 8 );
#endif
    COMPILE_ASSERT( sizeof(double) == 8 );

    unsigned mask    = DefaultTestMask;
    int      arg     = 1;
    bool     wasCpu  = false;
    bool     wasTest = false;
    bool     merge   = false;
    const char *mergeFile;

    if ( argc < 2 )
    {
      printUsage();
      return 1;
    }

    while ( arg < argc && argv[arg][0] == '-' )
    {                                       // read one "-*" argument

        if ( argv[arg][1] == 'b' ) {        // brief mode
            verbose = false;
            if ( argv[arg][2] == 'b' ) superBrief = true;
            }
        else

        if ( argv[arg][1] == 'h' ) {        // help
            printUsage();
            return 1;
            }
        else

        if ( argv[arg][1] == 'l' ) {        // listing
            listing = true;
            if ( argv[arg][2] == 's' ) spreadSheet = true;
            else
              if ( isdigit(argv[arg][2]) )
              {
                width = atoi(argv[arg]+2);
                if ( width > 160 ) width = 160;
                if ( width <  40 ) width = 40;
              }
            }
        else

        if ( argv[arg][1] == 's' ) {        // sort result table
            if ( argv[arg][2] == 'a' || argv[arg][2] == 'd' ) {
                ascending = (argv[arg][2] == 'a');
                sort = true;
                switch ( argv[arg][3] ) {
                    case 't':
                        sortType = SortTime;
                        break;
                    case 'r':
                        sortType = SortRef;
                        break;
                    case 'c':
                        sortType = SortClock;
                        break;
                    default:
                        sort = false;
                    }
                }
            else {
                sort = false;
                fprintf(stderr,"Error in sorting argument \"%s\" (will be ignored)\n",argv[arg]);
                }
            }
        else

        if ( argv[arg][1] == 'r' ) {        // test restrictions / regular expression
            char type = argv[arg][2];
            if ( !type ) {                  // regular expression
                if ( ++arg < argc )
                    restrictExpr = new VRegularExpression(argv[arg]);
                }
            else
            if ( type == 'm' || type == 'v' || type == 'd' ) {
                if ( ++arg < argc )
                    switch ( type ) {
                        case 'm':
                            restrictMemory = readBytes(argv[arg]);
                            break;
                        case 'v':
                            restrictVirtual = readBytes(argv[arg]);
                            break;
                        case 'd':
                            restrictDisk = readBytes(argv[arg]);
                            break;
                        }
                }
            else
                fprintf(stderr,"Error in restricting argument \"%s\" (will be ignored)\n",argv[arg]);
            }
        else

        if ( argv[arg][1] == 'i' )          // individual test
        {
          if ( ++arg < argc )
          {
            wasTest = true;
            int testNo = atoi( argv[arg] );
            if ( testNo < 1 ) testNo = 1;
            if ( testNo > MaxTest ) testNo = MaxTest;
            mask = 1 << (testNo-1);
            }
        }
        else

        if ( argv[arg][1] == 't' )          // self-test
        {
          RandomJames rnd;
          printHeader();
          printf( "RandomJames pseudo-random generator test: %s\n",
                  rnd.test() ? "passed" : "failed" );
          Sha1 sha;
          printf( "SHA-1 digest test: %s\n",
                  sha.test() ? "passed" : "failed" );
          return 0;
        }
        else

        if ( argv[arg][1] == 'm' )          // merge result files
        {
          if ( ++arg < argc )
          {
            merge = true;
            mergeFile = argv[arg];
          }
        }
        else

        if ( argv[arg][1] == 'c' )          // clock and CPU
        {
          if ( ++arg < argc )
          {
            wasCpu = true;
            actClock = atoi( argv[arg] );
            if ( arg+1 < argc )
              strcpy( actCpu = cpuBuf, argv[++arg] );
          }
        }
        else

        if ( argv[arg][1] == 'a' )          // log all results
          logAll = true;
        else

          fprintf( stderr, "Unknown argument \"%s\" (will be ignored), type \"bench -h\" for help!\n",
                   argv[arg] );

        arg++;
    }

    if ( verbose ) printHeader();

    if ( arg < argc && !wasTest )
        mask = (unsigned)strtol(argv[arg],NULL,0) & MaxTestMask;

    if ( !wasCpu ) {                        // read CPU-info from CPUFile
        FILE *f = fopen(CPUFile,"rt");
        if ( f ) {
            fscanf(f,"%*s%d",&actClock);    // skip the e-mail
            fgets(cpuBuf,MaxProcName,f);
            while ( cpuBuf[strlen(cpuBuf)-1] == '\n' ) cpuBuf[strlen(cpuBuf)-1] = (char)0;
            actCpu = cpuBuf;
            while ( isspace(*actCpu) ) actCpu++;
            fclose(f);
            }
        else {
            fprintf(stderr,"CPU type & frequency were not specified in the command line,\n"
                           "CPU file \"%s\" was not found => using default values [\"%s\",%d]!\n",
                    CPUFile,actCpu,actClock);
            }
        }

    if ( !readData() )
        fprintf(stderr,"Error reading old result file \"%s\"!\n",DataFile);

    if ( merge ) {                          // merge old result file
        if ( readData(mergeFile) ) {
            unifyData();
            writeData();
            printf( "OK merging the \"%s\" file. %d items were enhanced.\n",
                   mergeFile,enhancedCount);
            }
        else
            fprintf(stderr,"Cannot merge the \"%s\" file!\n",mergeFile);
        return 0;
        }

    if ( listing ) {                        // print actual results only
        arg = 0;
        while ( (1 << arg) & MaxTestMask ) {
            if ( (1 << arg) & mask ) listResult(arg+1);
            arg++;
            }
        return 0;
        }

        // test restrictions
    for ( arg = 0; (1 << arg) & MaxTestMask; arg++ )
        if ( ((1 << arg) & mask) &&
             (restrictMemory  && restrictMemory  < memoryLimits[arg] ||
              restrictVirtual && restrictVirtual < virtualLimits[arg] ||
              restrictDisk    && restrictDisk    < diskLimits[arg]) )
            mask -= (1 << arg);

    printf( "Running tests \"%04X\" on \"%s\" (%d MHz):\n", mask, actCpu, actClock );
    int elapsed;
    completion = 0;

        // 1 - eratosthen:
    if ( mask & 0x000001 )
    {
      elapsed = eratosthen( PrimeOrder, PrimeAssert );
      newResult( 1, elapsed );
    }

        // image transposition:
    if ( mask & 0x000002 )                  // 2 - small image
    {
      elapsed = imageTransposition( ImageWidth1, ImageHeight1, ImageIter1 );
      if ( elapsed ) newResult( 2, elapsed );
      else           notEnoughMemory( (ImageWidth1*ImageHeight1)>>19 );
    }

    if ( mask & 0x000004 )                  // 3 - bigger image
    {
      elapsed = imageTransposition( ImageWidth2, ImageHeight2, ImageIter2 );
      if ( elapsed ) newResult( 3, elapsed );
      else           notEnoughMemory( (ImageWidth2*ImageHeight2)>>19 );
    }

        // 4 - arbitrage sequence:
    if ( mask & 0x000008 )
    {
      startTimer();
      initInstance();
      solveInstance();
      elapsed = elapsedTime();
      newResult( 4, elapsed );
    }

        // 5 - needle throwing simulation:
    if ( mask & 0x000010 )
    {
      elapsed = needleSimulation();
      newResult( 5, elapsed );
    }

        // 6 - fast mem-copy:
    if ( mask & 0x000020 )
    {
      elapsed = memCopy();
      if ( elapsed ) newResult( 6, elapsed );
      else           notEnoughMemory( MemCopyMax>>19 );
    }

        // 7 - Monte-Carlo form-factor:
    if ( mask & 0x000040 )
    {
      elapsed = formFactor();
      newResult( 7, elapsed );
    }

        // 8 - Merge-sort on disk:
    if ( mask & 0x000080 )
    {
      elapsed = diskSort( 0 );
      newResult( 8, elapsed );
    }

        // 9 - T-S wavelet transform in memory:
    if ( mask & 0x000100 )
    {
      elapsed = waveletMemory();
      if ( elapsed ) newResult( 9, elapsed );
      else           notEnoughMemory( (WaveMaxSize*sizeof(int32))>>20 );
    }

        // 10 - T-S wavelet transform on disk:
    if ( mask & 0x000200 )
    {
      elapsed = waveletDisk( 0 );
      newResult( 10, elapsed );
    }

        // 11 - Adaptive K-D tree on disk:
    if ( mask & 0x000400 )
    {
      elapsed = akdTree( 0 );
      newResult( 11, elapsed );
    }

        // 12 - Quick-sort in virtual memory (short array /64MB/):
    if ( mask & 0x000800 )
    {
      elapsed = quickSortTest( QuickSortLen1 );
      if ( elapsed ) newResult( 12, elapsed );
      else           notEnoughMemory( QuickSortLen1>>17 );
    }

        // 13 - Quick-sort in virtual memory (middle array /256MB/):
    if ( mask & 0x001000 )
    {
      elapsed = quickSortTest( QuickSortLen2 );
      if ( elapsed ) newResult( 13, elapsed );
      else           notEnoughMemory( QuickSortLen2>>17 );
    }

        // 14 - Quick-sort in virtual memory (large array /1GB/):
    if ( mask & 0x002000 )
    {
      elapsed = quickSortTest( QuickSortLen3 );
      if ( elapsed ) newResult( 14, elapsed );
      else           notEnoughMemory( QuickSortLen3>>17 );
    }

        // 15 - Garbage collection test in 16MB of VM:
    if ( mask & 0x004000 )
    {
      elapsed = gcTest();
      newResult( 15, elapsed );
    }

        // 16 - Regular expression test:
    if ( mask & 0x008000 )
    {
      elapsed = regularTest();
      newResult( 16, elapsed );
    }

        // image transposition:
    if ( mask & 0x010000 )                  // 17 - small image
    {
      elapsed = imageTranspositionCorrect( ImageWidth1, ImageHeight1, ImageIter1 );
      if ( elapsed ) newResult( 17, elapsed );
      else           notEnoughMemory( (ImageWidth1*ImageHeight1)>>19 );
    }

    if ( mask & 0x020000 )                  // 18 - bigger image
    {
      elapsed = imageTranspositionCorrect( ImageWidth2, ImageHeight2, ImageIter2 );
      if ( elapsed ) newResult( 18, elapsed );
      else           notEnoughMemory( (ImageWidth2*ImageHeight2)>>19 );
    }

        // 19 - Merge-sort on disk (big version):
    if ( mask & 0x040000 )
    {
      elapsed = diskSort( 1 );
      newResult( 19, elapsed );
    }

        // 20 - T-S wavelet transform on disk (big version):
    if ( mask & 0x080000 )
    {
      elapsed = waveletDisk( 1 );
      newResult( 20, elapsed );
    }

        // 21 - Adaptive K-D tree on disk (big version):
    if ( mask & 0x100000 )
    {
      elapsed = akdTree( 1 );
      newResult( 21, elapsed );
    }

        // 22 - Parallel SHA-1 digest:
    if ( mask & 0x200000 )
    {
      elapsed = parallelSha( ShaMaxThreads );
      if ( elapsed ) newResult( 22, elapsed );
    }

        // 23 - Ctx-switch test - parallel SHA-1 digest:
    if ( mask & 0x400000 )
    {
      elapsed = parallelSha23();
      if ( elapsed ) newResult( 23, elapsed );
    }

        // 24 - Parallel sort in virtual memory:
    if ( mask & 0x800000 )
    {
      elapsed = parallelSort( PsMaxThreads );
      if ( elapsed ) newResult( 24, elapsed );
      else           notEnoughMemory( ParallelSortLen>>16 );
    }

    return completion;
}
