/*
    multithread.cc: multi-thread support for VUCAKO Bench.

    $Rev: 93 $
    $Date: 2007-03-17 16:25:56 +0100 (so, 17 III 2007) $

    Copyright (C) 2001-2007 by Josef Pelikan, MFF UK Prague
        http://cgg.ms.mff.cuni.cz/~pepca/

    Parts adopted from MSDK (C) 2004, Zlatko Michailov (LGPL license)

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pch.h"
#include "multithread.h"

//-----------------------------------------------------------
// Portable Critical section

#ifndef WIN32
pthread_mutex_t mutexInit = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#endif

CriticalSection::CriticalSection ()
{
#ifdef WIN32
  InitializeCriticalSection( &cs );
#else
  mutex = mutexInit;
#endif
  error = false;
}

void CriticalSection::enter () const
{
#ifdef WIN32
  EnterCriticalSection( &cs );
#else
  error = (pthread_mutex_lock( &mutex ) != 0);
#endif
}

bool CriticalSection::tryEnter () const
{
#ifdef WIN32
# ifdef _WINNT
  return( TryEnterCriticalSection( &cs ) != 0 );
# else
  EnterCriticalSection( &cs );
  return true;
# endif
#else
  error = false;
  return( pthread_mutex_trylock( &mutex ) == 0 );
#endif
}

void CriticalSection::leave () const
{
#ifdef WIN32
  LeaveCriticalSection( &cs );
#else
  error = (pthread_mutex_unlock( &mutex ) != 0);
#endif
}

CriticalSection::~CriticalSection ()
{
#ifdef WIN32
  DeleteCriticalSection( &cs );
#else
  error = (pthread_mutex_destroy( &mutex ) != 0);
#endif
}

//-----------------------------------------------------------
// Portable Mutex

Mutex::Mutex ()
{
#ifdef WIN32
  handle = CreateMutex( NULL, FALSE, NULL );
#else
  pthread_mutexattr_t attr;
  pthread_mutexattr_init( &attr );
  pthread_mutexattr_settype( &attr, PTHREAD_MUTEX_RECURSIVE_NP );
  pthread_mutex_init( &handle, &attr );
#endif
}

Mutex::~Mutex ()
{
#ifdef WIN32
  CloseHandle( handle );
#else
  pthread_mutex_destroy( &handle );
#endif
}

bool Mutex::lock ()
{
  bool ok = false;
#ifdef WIN32
  ok = (WaitForSingleObject( handle, INFINITE ) == WAIT_OBJECT_0);
#else
  ok = !pthread_mutex_lock( &handle );
#endif
  return ok;
}

bool Mutex::unlock ()
{
  bool ok = false;
#ifdef WIN32
  ok = (ReleaseMutex( handle ) != FALSE);
#else
  ok = !pthread_mutex_unlock( &handle );
#endif
  return ok;
}

//-----------------------------------------------------------
// Portable Semaphore

Semaphore::Semaphore ( long init, long maxCount )
{
  if ( init < 0L ) init = 0L;
#ifdef WIN32
  handle = CreateSemaphore( NULL, init, maxCount, NULL );
  error = (handle == NULL);
#else
  error = (sem_init( &sem, 0, (unsigned)init ) != 0);
    // maxCount value is ignored in pthreads..
#endif
}

void Semaphore::wait ()
{
#ifdef WIN32
  if ( handle != NULL )
    WaitForSingleObject( handle, INFINITE );
#else
  sem_wait( &sem );
#endif
}

bool Semaphore::tryWait ()
{
#ifdef WIN32
  if ( handle == NULL ) return false;
  return( WaitForSingleObject( handle, 0L ) == WAIT_OBJECT_0 );
#else
  return( sem_trywait( &sem ) == 0 );
#endif
}

void Semaphore::signal ( long count )
{
  if ( count < 1L ) count = 1L;
#ifdef WIN32
  if ( handle == NULL )
  {
    error = true;
    return;
  }
  error = (ReleaseSemaphore( handle, count, NULL ) != 0);
#else
  error = false;
  while ( !error && count-- > 0L )
    error = (sem_post( &sem ) != 0);
#endif
}

long Semaphore::getValue ()
{
#ifdef WIN32
  error = true;       // !!! not implemented in Win32 ?
  return 0L;
#else
#  ifdef __CYGWIN__
  error = true;
  return 0L;
#  else
  int val = 0;
  error = (sem_getvalue( &sem, &val ) != 0);
  return (long)val;
#  endif
#endif
}

Semaphore::~Semaphore ()
{
#ifdef WIN32
  if ( handle != NULL )
  {
    CloseHandle( handle );
    handle = NULL;
  }
#else
  sem_destroy( &sem );
#endif
}

//-----------------------------------------------------------
// Portable Thread

Thread::Thread ( void* _initData, int _stopTimeout )
{
  reset( _initData, true, _stopTimeout );
}

Thread::~Thread ()
{
  stop();
}

void Thread::_cleanup ( void* thisVoid )
{
  Thread* _this = (Thread*)thisVoid;

  _this->shutdown();
  _this->reset();
}

ThreadProcReturn THREADPROCCALL Thread::_exec ( void* thisVoid )
{
  Thread* _this = (Thread*)thisVoid;

#ifndef WIN32
  // Defferring cancelation must remain the first call to the pthread API.
  pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, 0 );
  pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, 0 );
  pthread_cleanup_push( _cleanup, thisVoid );
#endif

  // Hold startup until creation is complete
  _this->creation.wait();

  // Invoke startup() from here, not from run(). If run() is
  // overriden, startup() should still be invoked.
  _this->startup();

  // The thread routine itself.
  _this->run();

#ifdef WIN32
  // There is no stack under Windows, therefore invoke _cleanup() manually
  _cleanup( thisVoid );
#else
  // If execution comes to this point, run() finished gracefully,
  // and the _cleanup must be removed from the stack WITH execution.
  // That will invoke shutdown()
  pthread_cleanup_pop( 1 );
#endif

  return 0;
}

void Thread::reset ( void* _initData, bool withForce, int _stopTimeout )
{
  isRunning   = false;
  handle      = 0;
  id          = 0;
  stopTimeout = _stopTimeout;

  // Reset init data only if force is applied
  if ( withForce )
    initData = _initData;
}

bool Thread::init ( void* _initData, int _stopTimeout )
{
  bool ok = stop();
  if ( ok )
    reset( _initData, true, _stopTimeout );
  return ok;
}

bool Thread::start ()
{
  if ( !isRunning )
  {
#ifdef WIN32
    unsigned int tmpID;
    handle = (ThreadHandle)_beginthreadex( 0, 0, _exec, this, 0, &tmpID );
    id = (ThreadID)tmpID;
    isRunning = !!handle;
#else
    pthread_attr_t attr;
    pthread_attr_init( &attr );
    isRunning = !pthread_create( (PThreadHandle)&handle, &attr, _exec, this );
    id = handle;
    pthread_attr_destroy( &attr );
#endif

    // Now the new thread may run
    if ( isRunning )
      creation.signal();
  }

  return isRunning;
}

bool Thread::stop ()
{
  if ( isRunning )
  {
    isRunning = false;
#ifdef WIN32
    if ( WaitForSingleObject( handle, stopTimeout * 1000 ) == WAIT_TIMEOUT )
      terminate();
#else
    pthread_cancel( handle );
#endif
  }

  return true;
}

bool Thread::terminate ()
{
  bool ok = false;

#ifdef WIN32
  ok = !!TerminateThread( handle, (DWORD)-1 );
#else
  ok = !pthread_kill( handle, SIGKILL );
#endif

  isRunning = false;

  return ok;
}

bool Thread::waitUntilFinished ()
{
  bool ok = !isRunning;

  if ( !ok )
  {
#ifdef WIN32
    ok = (WaitForSingleObject( handle, INFINITE ) == WAIT_OBJECT_0);
#else
    ok = !pthread_join( handle, 0 );
#endif
  }

  return ok;
}
