/*
    multithread.h: multi-thread support for VUCAKO Bench.

    $Rev: 90 $
    $Date: 2007-03-13 10:46:11 +0100 (út, 13 III 2007) $

    Copyright (C) 2001-2007 by Josef Pelikan, MFF UK Prague
        http://cgg.ms.mff.cuni.cz/~pepca/

    Parts adopted from MSDK (C) 2004, Zlatko Michailov (LGPL license)

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef _MULTITHREAD_H
#define _MULTITHREAD_H

#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <memory.h>
#include <time.h>
#include <limits.h>

#ifdef WIN32
# include <windows.h>
# include <process.h>
# include <direct.h>
# include <io.h>
#else
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/time.h>
# include <unistd.h>
# include <pthread.h>
# include <semaphore.h>
#endif

//-----------------------------------------------------------
// Portable Critical section

class CriticalSection
{
protected:

#ifdef WIN32
  /** Win32 structure to represent state of a critical section. */
  mutable CRITICAL_SECTION cs;
#else
  /* POSIX mutex structure. */
  mutable pthread_mutex_t mutex;
#endif

public:

  /** error status (<code>true</code> should be very rare). */
  mutable bool error;

  /**
   * Initializes the critical section.
   * <p>The critical section can be entered recursively by the same thread..
   */
  CriticalSection ();

  /**
   * Enters (locks) the critical section. If the section is already locked, suspends (blocks)
   * the current thread until it is unlocked again.
   * <p>One process is able to call 'enter()' more than one times.
   */
  void enter () const;

  /**
   * Test if the critical section is locked - if not, enters it immediately and returns <code>true</code>.
   * Otherwise returns <code>false</code> (does not block in any case).
   *
   * @return <code>true</code> if (non-blocking) enter operation was executed successfully.
   */
  bool tryEnter () const;

  /**
   * Leaves (unlocks) previously entered critical section.
   * <p>Number of 'enter()' (plus successfull 'tryEnter()') operations
   * must match with number of subsequent 'leave()' calls!
   */
  void leave () const;

  /**
   * Destroys the critical section.
   */
  ~CriticalSection ();

};

//-----------------------------------------------------------
// Portable Mutex

class Mutex
{
protected:

#ifdef WIN32
  /** Internal mutex handle. */
  HANDLE handle;
#else
  /** Internal mutex handle. */
  pthread_mutex_t handle;
#endif

public:

  /** Creates mutex object without obtaining it. */
  Mutex ();

  /** Releases the lock and destroys the object. */
  ~Mutex ();

  /**
   * Waits indefinitely for mutex lock.
   * Returns true if succeeded.
   */
  bool lock ();

  /**
   * Releases previously obtained mutex lock.
   * Returns true if succeeded.
   */
  bool unlock ();
};

//-----------------------------------------------------------
// Portable Semaphore

class Semaphore
{
protected:

#ifdef WIN32
  /** HANDLE returned by ::CreateSemaphore() call */
  HANDLE handle;
#else
  /** POSIX semaphore structure. */
  sem_t sem;
#endif

public:

  /** error status (<code>true</code> should be very rare). */
  bool error;

  /**
   * Initializes the semaphore.
   * <p>#error is set if maximum allowed counter value is exceeded.
   * Generally - failure reasons are not portable.
   *
   * @param  init Initial counter value (must be <code>&gt;=0</code>).
   * @param  maxCount Maximum counter value (=maximum number of shared objects/titbits/signals).
   *         Not supported under PTHREADS.
   */
  Semaphore ( long init =0L, long maxCount =LONG_MAX );

  /**
   * Wait operation.
   * Suspends (blocks) the current thread until the semaphore has positive counter value.
   * <p>#error variable is not affected.
   */
  void wait ();

  /**
   * Conditional wait operation.
   * Test if the semaphore's counter is positive. Returns <code>true</code> if
   * it has succeeded, <code>false</code> if the semaphore is blocked.
   * Returns immediately in both cases.
   * <p>#error variable is not affected.
   *
   * @return <code>true</code> if (non-blocking) wait operation was executed successfully
   */
  bool tryWait ();

  /**
   * Signal operation.
   * Increments the semaphore's counter and can cause re-scheduling of some thread[s]
   * waiting for that..
   *
   * @param  count Increment count (number of potentially waiting threads to start again).
   *               If maximum allowed counter value is exceeded,
   *               #error is set.
   */
  void signal ( long count =1L );

  /**
   * Gets the actual semaphore counter value (non-mandatory, not supported under PTHREADS).
   *
   * @return Actual counter value.
   */
  long getValue ();

  /**
   * Destroys the semaphore instance.
   */
  ~Semaphore ();

};

//-----------------------------------------------------------
// Portable Thread

#ifdef WIN32
  typedef HANDLE ThreadHandle;
  typedef ThreadHandle PThreadHandle;
  typedef DWORD ThreadID;
  typedef unsigned int ThreadProcReturn;
# define THREADPROCCALL __stdcall
#else
  typedef pthread_t ThreadHandle;
  typedef ThreadHandle* PThreadHandle;
  typedef pthread_t ThreadID;
  typedef void* ThreadProcReturn;
# define THREADPROCCALL
#endif

const int DEFAULT_STOP_TIMEOUT = 30;

/**
 * Thread class - subclass it and override the 'run()' method.
 */
class Thread
{
  //-----------------------------------------------------------
  // data:

protected:

  Semaphore creation;
  volatile bool isRunning;
  volatile ThreadHandle handle;
  volatile ThreadID id;
  volatile void* initData;
  volatile int stopTimeout;

  //-----------------------------------------------------------
  // support (non-overridable) methods:

private:

  /**
   * This method is invoked when 'pthread_cancel()' is called.
   * It forwards the execution to the 'shutdown()' overridable.
   */
  static void _cleanup ( void* thisVoid );

  /** Low-level thread proc. */
  static ThreadProcReturn THREADPROCCALL _exec ( void* thisVoid );

  //-----------------------------------------------------------
  // support (overridable) methods:

protected:

  void reset ( void* _initData =NULL, bool withForce =false, int _stopTimeout =DEFAULT_STOP_TIMEOUT );

  //-----------------------------------------------------------
  // public access methods:

public:

  /**
   * Initializes the data members without creating an execution thread.
   */
  Thread ( void* _initData =NULL, int _stopTimeout =DEFAULT_STOP_TIMEOUT );

  virtual ~Thread ();

  /**
   * This method may be called from the creating thread after the
   * new thread instance is constructed and before 'start()' is called.
   * The purpose is to allow the new thread instance to initialize itself
   * by overriding this method.
   */
  virtual bool init ( void* _initData =NULL, int _stopTimeout =DEFAULT_STOP_TIMEOUT );

  /** Starts the execution. */
  bool start ();

  /**
   * Attempts to stop the execution thread gracefully
   * Windows: allow stopTimeout seconds for the !isRunning
   * to be processed. Then brutally terminate the thread.
   */
  bool stop ();

  /** Terminates the execution thread brutally. */
  bool terminate ();

  /** Waits for regular thread-proc finish. */
  bool waitUntilFinished ();

  ThreadID getThreadID ()
  {
    return id;
  }

  bool isThreadRunning ()
  {
    return isRunning;
  }

  //-----------------------------------------------------------
  // overridables:

protected:

  /** This method is invoked on behalf of the new thread before 'run()'. */
  virtual void startup ()
  {}

  /** This method is invoked on behalf of the dying thread after 'run()'. */
  virtual void shutdown ()
  {}

  /**
   * This represents the main thread proc of the thread. It is invoked
   * after 'startup()' and before 'shutdown()'.
   */
  virtual void run () =0;

};

#endif
