/*
  sha1.cc

  SHA-1 digest in C by Steve Reid <steve@edmweb.com> (100% Public Domain).
  Adopted by Josef Pelikan.

  $Rev: 93 $
  $Date: 2007-03-17 16:25:56 +0100 (so, 17 III 2007) $

  Test Vectors (from FIPS PUB 180-1):

  "abc"
  A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D

  "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
  84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1

  A million repetitions of "a"
  34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
*/

#include "pch.h"
#include "sha1.h"

//-----------------------------------------------------------
// Portable Critical section

#define rol(value,bits) (((value)<<(bits))|((value)>>(32-(bits))))

// blk0() and blk() perform the initial expand.
// I got the idea of expanding during the round function from SSLeay
#ifdef HAVE_BIG_ENDIAN
# define blk0(i) l[i]
#else
# define blk0(i) (l[i] = (rol(l[i],24)&0xFF00FF00)|(rol(l[i],8)&0x00FF00FF))
#endif

#define blk(i) (l[i&15] = rol(l[(i+13)&15]^l[(i+8)&15]^l[(i+2)&15]^l[i&15],1))

// (R0+R1), R2, R3, R4 are the different operations used in SHA1
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);

// Hash a single 512-bit block. This is the core of the algorithm
void Sha1::transform ()
{
    // Copy context->state[] to working vars
  unsigned32 a = state[0];
  unsigned32 b = state[1];
  unsigned32 c = state[2];
  unsigned32 d = state[3];
  unsigned32 e = state[4];

    // 4 rounds of 20 operations each. Loop unrolled
  R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
  R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
  R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
  R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
  R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
  R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
  R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
  R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
  R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
  R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
  R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
  R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
  R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
  R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
  R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
  R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
  R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
  R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
  R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
  R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);

    // Add the working vars back into state[]
  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;
  state[4] += e;
}

Sha1::Sha1 ()
{
  init();
}

void Sha1::init ()
{
    // SHA1 initialization constants
  state[0] = 0x67452301;
  state[1] = 0xEFCDAB89;
  state[2] = 0x98BADCFE;
  state[3] = 0x10325476;
  state[4] = 0xC3D2E1F0;
  count[0] =
  count[1] = 0;
}

void Sha1::update ( const byte *data, unsigned len )
{
  unsigned i, j;

  j = (count[0] >> 3) & 63;
  if ( (count[0] += len << 3) < (len << 3) ) count[1]++;
  count[1] += (len >> 29);

  if ( (j + len) > 63 )
  {
    memcpy( c + j, data, (i = 64 - j) );
    transform();
    for ( ; i + 63 < len; i += 64 )
    {
      memcpy( c, data + i, 64 );
      transform();
    }
    j = 0;
  }
  else
    i = 0;

  memcpy( c + j, data + i, len - i );
}

void Sha1::finish ( byte *digest )
{
  unsigned i;
  byte finalcount[8];

  for ( i = 0; i < 8; i++ )
    finalcount[i] = (byte)((count[(i >= 4 ? 0 : 1)] >> ((3 - (i & 3)) * 8) ) & 255);
                                            // Endian independent

  update( (byte*)"\200", 1 );
  while ( (count[0] & 504) != 448 )
    update( (byte*)"\0", 1 );

  update( finalcount, 8 );                  // Should cause a transform()

  for ( i = 0; i < 20; i++ )
    digest[i] = (byte)((state[i>>2] >> ((3 - (i & 3)) * 8) ) & 255);
}

bool Sha1::test ()
{
  const byte OK_ABC[] =
  {
    0xA9, 0x99, 0x3E, 0x36,
    0x47, 0x06, 0x81, 0x6A,
    0xBA, 0x3E, 0x25, 0x71,
    0x78, 0x50, 0xC2, 0x6C,
    0x9C, 0xD0, 0xD8, 0x9D
  };
  const byte OK_ABCDBC[] =
  {
    0x84, 0x98, 0x3E, 0x44,
    0x1C, 0x3B, 0xD2, 0x6E,
    0xBA, 0xAE, 0x4A, 0xA1,
    0xF9, 0x51, 0x29, 0xE5,
    0xE5, 0x46, 0x70, 0xF1
  };
  const byte OK_1MA[] =
  {
    0x34, 0xAA, 0x97, 0x3C,
    0xD4, 0xC4, 0xDA, 0xA4,
    0xF6, 0x1E, 0xEB, 0x2B,
    0xDB, 0xAD, 0x27, 0x31,
    0x65, 0x34, 0x01, 0x6F
  };

  byte digest[ 20 ];

  init();
  update( (const byte*)"abc", 3 );
  finish( digest );
  if ( memcmp( digest, OK_ABC, 20 ) ) return false;

  init();
  update( (const byte*)"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", 56 );
  finish( digest );
  if ( memcmp( digest, OK_ABCDBC, 20 ) ) return false;

  init();
  for ( int i = 0; i++ < 100000; )
    update( (const byte*)"aaaaaaaaaa", 10 );
  finish( digest );
  if ( memcmp( digest, OK_1MA, 20 ) ) return false;

  return true;
}
