/*
    test.cc: VUCAKO Bench multi-thread core (test routines).

    $Rev: 93 $
    $Date: 2007-03-17 16:25:56 +0100 (so, 17 III 2007) $

    Copyright (C) 2001-2007 by Josef Pelikan, MFF UK Prague
        http://cgg.ms.mff.cuni.cz/~pepca/

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pch.h"

//-----------------------------------------------------------
//  Parallel sort in virtual memory:

const int PsCells = ParallelSortLen / PsCellSize;

/** One instance of worker-task for parallel sorting. */
struct PsTask
{
  /** -1 .. unused, 0 .. initial quick-sort, 1 .. merge. */
  int type;

  /** Data address in cell-index. */
  int addr;

  /** Destination data size in array items. */
  int size;

  /** Log(size) = 1 .. finished initial sorting (PsCellSize), .. */
  int logSize;
};

/** Control data structure for parallel sorting. */
struct PsControl
{
  /** Task queue & finish protection. */
  CriticalSection cs;

  /** Worker threads are running? */
  volatile bool running;

  /** Finish-signaling object. */
  Semaphore fin;

  /** Semaphore to control access to worker tasks (signal => a task is ready). */
  Semaphore avail;

  /** Task queue. */
  PsTask tasks[ PsCells ];

  /** Put index (item index to be putted next time). */
  int putPtr;

  /** Get index (item index to be looked at next time). */
  int getPtr;

  /** Array to be sorted - primary and secondary one. */
  double *data[ 2 ];

  /** Which cells are finished? (log(size)) */
  byte finish[ PsCells ];

  /**
   * Puts new computing task in the processing queue.
   * Critical section has to be locked!
   */
  void newTask ( PsTask &task )
  {
      // I'm sure there is free space in the tasks[] array..
    tasks[putPtr++] = task;
    if ( putPtr >= PsCells ) putPtr = 0;
    cs.leave();

      // release another task..
    avail.signal();
  }

  /** Initial-sort task has been finished .. should we start new merge task? */
  void finishSort ( PsTask &task )
  {
    cs.enter();
      // check my sibling..
    finish[ task.addr ] = 1;
    if ( finish[ task.addr ^ 1 ] != 1 )
    {
      cs.leave();
      return;
    }

      // prepare new merge-task:
    task.type  = 1;
    task.addr &= -2;
    task.size  = PsCellSize + PsCellSize;
    task.logSize++;

    newTask( task );
  }

  /** Merge task has been finished .. should we start new (bigger) one? */
  void finishMerge ( PsTask &task )
  {
      // was I the last merge-task?
    int size = task.size;
    if ( size == ParallelSortLen )
    {
      fin.signal();
      return;
    }

    int cells   = size / PsCellSize;
    int sibling = (task.addr ^ cells);

    cs.enter();
      // check if my sibling is ready to merge with me..
    finish[ task.addr ] = task.logSize;
    if ( finish[ sibling ] != task.logSize )
    {
      cs.leave();
      return;
    }

      // prepare new merge-task:
    if ( task.addr > sibling ) task.addr = sibling;
    task.size += task.size;
    task.logSize++;

    newTask( task );
  }

  void getTask ( PsTask &task )
  {
    cs.enter();
      // I'm sure there is at least one task available..
    task = tasks[getPtr];
    tasks[getPtr++].type = -1;
    if ( getPtr >= PsCells ) getPtr = 0;
    cs.leave();
  }

};

/** Worker-thread for parallel sorting. */
class PsThread : public Thread
{
protected:

  virtual void run ()
  {
    PsControl *c = (PsControl*)initData;

    do
    {
        // wait for a task to process..
      c->avail.wait();
      if ( !c->running ) return;

      PsTask task;
      c->getTask( task );

      if ( task.type == 0 )                 // initial sort
      {
        qs( c->data[0] + task.addr * PsCellSize, PsCellSize );
        c->finishSort( task );
      }

      else                                  // merge two already sorted segments
      {
        int cnt1 = (task.size >> 1);
        int cnt2 = task.addr * PsCellSize;
        double *in1 = c->data[ task.logSize & 1 ] + cnt2;
        double *in2 = in1 + cnt1;
        double *dst = c->data[ 1 - (task.logSize & 1) ] + cnt2;
        cnt2 = cnt1;

          // merge-loop:
        do
          if ( *in1 < *in2 )
          {
            *dst++ = *in1++;
            if ( --cnt1 <= 0 ) break;
          }
          else
          {
            *dst++ = *in2++;
            if ( --cnt2 <= 0 ) break;
          }
        while ( true );

          // append tail_1?
        while ( cnt1-- > 0 )
          *dst++ = *in1++;

          // append tail_2?
        while ( cnt2-- > 0 )
          *dst++ = *in2++;

        c->finishMerge( task );
      }

    } while ( true );
  }
};

/**
 * One run (the complete parallel sorting job utilizing the given
 * number of worker threads).
 */
int parallelSortRun ( int threads )
{
  startTimer();
  int elapsed = 0;

  PsControl c;
  c.running = true;
  memset( c.tasks, -1, sizeof( c.tasks ) );
  c.putPtr  =
  c.getPtr  = 0;
  memset( c.finish, 0, sizeof( c.finish ) );
  c.data[0] = (double*)VIRTUAL_ALLOC( ParallelSortLen * sizeof(double) );
  c.data[1] = (double*)VIRTUAL_ALLOC( ParallelSortLen * sizeof(double) );

  if ( c.data )
  {                                         // enough memory
      // 1. create worker threads:
    PsThread *workers = new PsThread[ threads ];
    int i;
    for ( i = 0; i < threads; i++ )
    {
      workers[i].init( &c );
      workers[i].start();
    }

      // 2. generate primary tasks:
    RandomJames rnd;
    double *ptr = c.data[0];
    int cell = 0;
    while ( cell < PsCells )
    {
      for ( i = 0; i++ < PsCellSize; )
        *ptr++ = rnd.uniformNumber();

      PsTask t;
      t.type    = 0;
      t.addr    = cell;
      t.size    = PsCellSize;
      t.logSize = 1;

      c.cs.enter();
      c.newTask( t );

      cell++;
    }

      // 3. wait until everything is finished:
    c.fin.wait();
    c.running = false;
    c.avail.signal( threads );              // stop all working threads..

    for ( i = 0; i < threads; i++ )
      workers[i].waitUntilFinished();

      // 4. check the result:
    ptr = c.data[0];
    for ( i = 1; i++ < ParallelSortLen; ptr++ )
      if ( ptr[0] > ptr[1] )
      {
        elapsed = -1;
        break;
      }

    delete[] workers;
    VIRTUAL_FREE( c.data[0], ParallelSortLen * sizeof(double) );
    VIRTUAL_FREE( c.data[1], ParallelSortLen * sizeof(double) );

    if ( !elapsed ) elapsed = elapsedTime();
  }

  return elapsed;
}

int parallelSort ( int maxThreads )
{
  int threads = 2;
  int elapsed = 0;

  do
  {
    int runElapsed = parallelSortRun( threads );

    if ( runElapsed <= 0 )
    {
      if ( runElapsed < 0 && !superBrief )
        printf( "Parallel-sort using %2d threads (double[%dM]): fail!\n",
                (ParallelSortLen>>20), threads );
      return 0;
    }

    if ( !superBrief )
    {
      printf( "Parallel-sort using %2d threads (double[%dM]): %.2f sec\n",
              threads, (ParallelSortLen>>20), runElapsed*0.001 );
      fflush( stdout );
    }

    elapsed += runElapsed;
  }
  while ( (threads += threads) <= maxThreads );

  if ( !superBrief )
    printf( "Parallel-sort using %d-%d threads (double[%dM]): %.2f sec total\n",
            2, maxThreads, (ParallelSortLen>>20), elapsed*0.001 );

  return elapsed;
}

//-----------------------------------------------------------
//  Parallel SHA-1 digest:

/** One instance of worker-task for SHA-1. */
class ShaTask
{
public:

  /** 0 .. generate data file, 1 .. SHA-1 digest */
  int type;

  /** Pseudo-random seed for task #0. */
  int seed;

  /** Data array. */
  byte *data;

  ShaTask ( int _seed, int len )
  {
    type = 0;
    seed = _seed;
    data = new byte[ len ];
  }

  ~ShaTask ()
  {
    delete[] data;
  }
};

/** Worker-thread for parallel SHA-1 digest. */
class ShaThread : public Thread
{
protected:

  virtual void run ();
};

/** Control data structure for parallel SHA-1 digest. */
struct ShaControl
{
  /** Task queue & result protection. */
  CriticalSection cs;

  /** Worker threads are running? */
  volatile bool running;

  /** Actual: data file size. */
  int dataSize;

  /** Actual: number of data files. */
  int dataFiles;

  /** Correct digest result. */
  const byte *okDigest;

  /** Computed total digest sum (for checking purpose only). */
  byte totalDigest[ 20 ];

  /** Semaphore to control access to worker tasks. */
  Semaphore *avail;

  /** Task queue semaphore. */
  Semaphore *queue;

  /** Task queue. */
  ShaTask *tasks[ ShaQueueSize ];

  /** Put index (item index to be putted next time). */
  int putPtr;

  /** Get index (item index to be looked at next time). */
  int getPtr;

  void putTask ( ShaTask *task )
  {
    cs.enter();

      // I'm sure there is free space in the tasks[] array..
    tasks[ putPtr++ ] = task;
    if ( putPtr >= ShaQueueSize ) putPtr = 0;

    cs.leave();
  }

  ShaTask *getTask ()
  {
    cs.enter();

      // I'm sure there is at least one task available..
    ShaTask *tmp      = tasks[ getPtr ];
    tasks[ getPtr++ ] = NULL;
    if ( getPtr >= ShaQueueSize ) getPtr = 0;

    cs.leave();
    return tmp;
  }

  void setResult ( const byte *digest )
  {
    byte *ptr = totalDigest;
    cs.enter();

    for ( int i = 0; i++ < 20; )
      *ptr++ ^= *digest++;

    cs.leave();
  }

  /**
   * One run (the complete SHA-1 computation utilizing the given
   * number of worker threads).
   */
  int oneRun ( int threads )
  {
    startTimer();

    running = true;
    memset( totalDigest, 0, sizeof(totalDigest) );
    memset( tasks,       0, sizeof(tasks) );
    putPtr  =
    getPtr  = 0;
    avail   = new Semaphore();
    queue   = new Semaphore( ShaQueueSize );

      // 1. create worker threads:
    ShaThread *workers = new ShaThread[ threads ];
    int i;
    for ( i = 0; i < threads; i++ )
    {
      workers[i].init( this );
      workers[i].start();
    }

      // 2. generate primary tasks:
    RandomJames rnd;
    for ( i = 0; i < dataFiles; i++ )
    {
      queue->wait();
      ShaTask *t = new ShaTask( (int)( rnd.uniformNumber() * 0x7fffffff ), dataSize );
      putTask( t );
      avail->signal();
    }

      // 3. wait until everything is finished:
    for ( i = 0; i++ < ShaQueueSize; )
      queue->wait();
    running = false;
    avail->signal( threads );               // stop all working threads..

    for ( i = 0; i < threads; i++ )
      workers[i].waitUntilFinished();

    delete[] workers;
    delete   queue;
    delete   avail;

      // 4. check the result:
    return( memcmp( totalDigest, okDigest, 20 ) ? 0 : elapsedTime() );
  }

};

void ShaThread::run ()
{
  ShaControl *c = (ShaControl*)initData;
  Sha1 sha;
  byte digest[ 20 ];

  do
  {
      // wait for a task to process..
    c->avail->wait();
    if ( !c->running ) return;

    ShaTask *task = c->getTask();

    if ( task->type == 0 )                // generate data file
    {
      int   seed = task->seed;
      byte *ptr  = task->data;
      int   i    = c->dataSize;
      while ( i-- )
      {
        *ptr++ = (byte)( (seed >> 19) & 0xff );
        seed   = seed * 149 + 10007;
      }
      task->type = 1;
      c->putTask( task );
      c->avail->signal();                 // process another task
    }

    else                                  // compute SHA-1 digest
    {
      sha.init();
      sha.update( task->data, c->dataSize );
      delete task;
      sha.finish( digest );
      c->setResult( digest );
      c->queue->signal();                 // making room for another task
    }
  }
  while ( true );
}

/** Correct total digest sum (#22). */
const byte OK_DIGEST[] =
{
  0xB3, 0x80, 0x98, 0x26,
  0x3E, 0x74, 0x9B, 0x74,
  0xAF, 0x6A, 0x36, 0xB6,
  0xED, 0x1D, 0x5C, 0xDD,
  0x69, 0xBD, 0x3F, 0x57
};

/** Correct total digest sum (#23). */
const byte OK_DIGEST2[] =
{
  0xC4, 0xF3, 0xF6, 0x51,
  0x0A, 0x3D, 0x76, 0x5A,
  0x16, 0xAC, 0x37, 0xB9,
  0x01, 0xE9, 0x1B, 0x39,
  0xD8, 0xB4, 0x1E, 0x85
};

int parallelSha ( int maxThreads )
{
  int threads = 1;
  int elapsed = 0;
  ShaControl c;
  c.dataFiles = ShaDataFiles;
  c.dataSize  = ShaDataSize;
  c.okDigest  = OK_DIGEST;

  do
  {
    int runElapsed = c.oneRun( threads );

    if ( runElapsed <= 0 )
    {
      printf( "Parallel SHA-1 digest using %2d threads: fail!\n", threads );
      for ( int i = 0; i < 20; i++ )
        printf( " %02X", (unsigned)c.totalDigest[i] );
      putchar( '\n' );
      return 0;
    }

    if ( !superBrief )
    {
      printf( "Parallel SHA-1 digest using %2d threads: %.2f sec\n",
              threads, runElapsed * 0.001 );
      fflush( stdout );
    }

    elapsed += runElapsed;
  }
  while ( (threads += threads) <= maxThreads );

  if ( !superBrief )
    printf( "Parallel SHA-1 digest using 1-%2d threads: %.2f sec total\n",
            maxThreads, elapsed * 0.001 );

  return elapsed;
}

int parallelSha23 ()
{
  ShaControl c;
  c.dataFiles = (ShaDataFiles << ShaShift23);
  c.dataSize  = (ShaDataSize  >> ShaShift23);
  c.okDigest  = OK_DIGEST2;

  int elapsed = c.oneRun( ShaThreads23 );

  if ( elapsed <= 0 )
  {
    printf( "Ctx-switch test (parallel SHA-1) using %2d threads: fail!\n", ShaThreads23 );
    for ( int i = 0; i < 20; i++ )
      printf( " %02X", (unsigned)c.totalDigest[i] );
    putchar( '\n' );
    return 0;
  }

  if ( !superBrief )
    printf( "Ctx-switch test (parallel SHA-1) using %2d threads: %.2f sec total\n",
            ShaThreads23, elapsed * 0.001 );

  return elapsed;
}
